"""
The ToolManager manages a list of tools, context, and all ways to execute functions.
"""

from typing import (
    List,
    Dict,
    Optional,
    Any,
    Callable,
    Awaitable,
    Union,
    overload,
    Literal,
)
import logging
from copy import deepcopy
import asyncio
from adalflow.optim.parameter import Parameter, ParameterType
import nest_asyncio
import warnings

from adalflow.core.container import ComponentList
from adalflow.optim.grad_component import GradComponent
from adalflow.core.component import Component
from adalflow.core.func_tool import FunctionTool
from adalflow.core.types import (
    FunctionDefinition,
    FunctionOutput,
    Function,
    FunctionExpression,
)
from adalflow.utils import printc


from adalflow.core.functional import (
    parse_function_call_expr,
    sandbox_exec,
)

log = logging.getLogger(__name__)


AsyncCallable = Callable[..., Awaitable[Any]]

ToolType = Union[FunctionTool, Callable[..., Any], Awaitable[Callable[..., Any]]]
ToolsType = List[ToolType]


def run_async_in_new_loop(coro):
    """Run async function in a new event loop."""
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        return loop.run_until_complete(coro)
    finally:
        loop.close()
        asyncio.set_event_loop(None)


class CallFunctionTool(Component):
    __doc__ = """Contains other unit gradcomponent such as calling
                a FunctionTool"""

    def __init__(self):
        super().__init__()

    def forward(self, func: Parameter, context: Dict[str, object]):
        return self.bicall(func, context=context)

    def call(self, func: Function, context: Dict[str, object]) -> FunctionOutput:
        return self.bicall(func, context=context)

    def bicall(
        self,
        func: Union[Function, Parameter],
        context: Dict[str, object] = {},
    ):
        if isinstance(func, Parameter):
            # printc(f"context: {context}", color="yellow")
            func_data: Function = func.map_to_successor(self)
            if not isinstance(func_data, Function):
                raise ValueError(f"Error parsing function expression: {func}")
            tool: FunctionTool = context[func_data.name]
            # print(f"tool training: {tool.training}")
            output = tool.forward(*func_data.args, **func_data.kwargs)

            from adalflow.optim.grad_component import fun_to_grad_component

            # this will automatically create the outputparam, and connect output, func to the outputParam
            @fun_to_grad_component()
            def func_to_funcoutput(output, func):
                return output

            # NOTE: for nontrainable function but the overall tool manager is in training mode,
            # we will create a similar func to output edge to handle the feedback backpapagation
            if not isinstance(output, Parameter):
                return func_to_funcoutput.forward(output, func)
            else:
                # reconnect the predecessor for tracing as it is not done in tool.forward
                output.predecessors.add(func)
            return output
        else:
            tool: FunctionTool = context[func.name]
            output = tool.call(*func.args, **func.kwargs)
            return output


class FunctionExperssionToFunction(GradComponent):
    def __init__(self):
        super().__init__(desc="Convert FunctionExpression to Function")

    def call(self, expr: FunctionExpression, context: Dict[str, object]) -> Function:

        assert isinstance(
            expr, FunctionExpression
        ), f"Expected FunctionExpression, got {type(expr)}"

        expr_str = expr.action
        func_name, args, kwargs = parse_function_call_expr(expr_str, context)

        output = Function(
            name=func_name,
            args=args,
            kwargs=kwargs,
            thought=expr.thought,
        )
        return output


# TODO: good to track all the failed function calls
# Tool manager is a task component
class ToolManager(Component):
    __doc__ = r""""Manage a list of tools, context, and all ways to execute functions.


    ToolManager is a task component that does not need its own backward function.

    yaml and json definitions are for quick access to the definitions of the tools.
    If you need more specification, such as using exclude field, you can use the function_definitions.
    """

    def __init__(
        self,
        tools: ToolsType = [],
        additional_context: Optional[
            Dict[str, object]
        ] = {},  # anything besides the tools
    ):
        super().__init__()
        nest_asyncio.apply()  # Apply nest_asyncio to handle nested loops
        processed_tools = [
            (
                FunctionTool(fn=deepcopy(tool))
                if not isinstance(tool, FunctionTool)
                else tool
            )
            for tool in tools
        ]

        self.tools = ComponentList(processed_tools)
        self._context_map = self.create_context_map_from_tools(self.tools)
        self._additional_context = additional_context or {}
        self.context = {**self._context_map, **self._additional_context}
        log.info(
            f"Initialized ToolManager with {len(self.tools)} tools and additional context {self._additional_context}"
        )

    @staticmethod
    def get_context_index(tool: FunctionTool) -> Dict[str, object]:
        index = tool.definition.func_name
        if tool.definition.class_instance:
            index = f"{tool.definition.class_instance}.{index}"
        output = {index: tool}
        if tool.definition.func_name == "__call__":
            # add another index of directly using the classinstance
            output[f"{tool.definition.class_instance}"] = tool
        return output

    @staticmethod
    def create_context_map_from_tools(tools: List[FunctionTool]) -> Dict[str, object]:
        output: Dict[str, object] = {}
        for tool in tools:
            tool_map = ToolManager.get_context_index(tool)
            for k, v in tool_map.items():
                if k in output:
                    # raise ValueError(f"Duplicate key {k} in the context map.")
                    warnings.warn(f"Duplicate key {k} in the context map.")
                    continue
                output[k] = v
        return output

    @property
    def yaml_definitions(self) -> List[str]:
        output = []
        for tool in self.tools:
            if not tool.definition.class_instance:
                output.append(tool.definition.to_yaml(exclude=["class_instance"]))
            else:
                output.append(tool.definition.to_yaml())
        return output

    @property
    def json_definitions(self) -> List[str]:
        output = []
        for tool in self.tools:
            if not tool.definition.class_instance:
                output.append(tool.definition.to_json(exclude=["class_instance"]))
            else:
                output.append(tool.definition.to_json())
            output.append(tool.definition.to_json(exclude=["class_instance"]))
        return output

    @property
    def function_definitions(self) -> List[FunctionDefinition]:
        return [tool.definition for tool in self.tools]

    def parse_func_expr(
        self,
        expr: Union[FunctionExpression, Parameter],
        map_fn: Callable = lambda x: x.data,
    ) -> Union[Function, Parameter]:
        r"""Parse the function call expression."""

        if isinstance(expr, Parameter):
            func = FunctionExperssionToFunction()
            expr.add_successor_map_fn(func, map_fn=map_fn)
            output = func.forward(expr, context=self.context)
            return output
        else:
            try:
                expr_str = expr.action
                func_name, args, kwargs = parse_function_call_expr(
                    expr_str, self.context
                )
                return Function(name=func_name, args=args, kwargs=kwargs)
            except Exception as e:
                log.error(f"Error {e} parsing function call expression: {expr}")
                raise ValueError(f"Error {e} parsing function call expression: {expr}")

    @overload
    def call(
        self, *, expr_or_fun: FunctionExpression, step: Literal["parse"] = "parse"
    ) -> Function: ...

    @overload
    def call(
        self, *, expr_or_fun: FunctionExpression, step: Literal["execute"] = "execute"
    ) -> FunctionOutput: ...

    @overload
    def call(
        self, *, expr_or_fun: Function, step: Literal["execute"] = "parse"
    ) -> Function: ...

    @overload
    def call(
        self, *, expr_or_fun: Function, step: Literal["execute"] = "execute"
    ) -> FunctionOutput: ...

    def call(
        self,
        *,
        expr_or_fun: Union[FunctionExpression, Function],
        step: Literal["execute"] = "execute",
    ) -> Union[FunctionOutput, Function, Parameter]:
        if not isinstance(expr_or_fun, (Function, FunctionExpression)):
            raise ValueError(
                f"expr_or_fun should be either a Function or FunctionExpression. Got {expr_or_fun}"
            )
        if step == "parse":
            if isinstance(expr_or_fun, Function):
                return expr_or_fun
            return self.parse_func_expr(expr_or_fun)
        elif step == "execute":
            if isinstance(expr_or_fun, Function):
                return self.execute_func(expr_or_fun)
            return self.execute_func_expr(expr_or_fun)
        else:
            raise ValueError(f"step should be either 'parse' or 'execute'. Got {step}")

    def forward(
        self,
        *,
        expr_or_fun: Union[FunctionExpression, Function, Parameter],
        step: Literal["parse", "execute"] = "execute",
        map_fn: Callable = lambda x: x.data,  # how to map the parameter to the needed data
    ) -> Union[FunctionOutput, Function, Parameter]:
        "Run a forward pass on the tool manager such as parsing function expression or executing function."
        if isinstance(expr_or_fun, Parameter):
            expr_or_fun_data = map_fn(expr_or_fun)
            if step == "execute":
                if isinstance(expr_or_fun_data, Function):
                    return self.execute_func(expr_or_fun, map_fn=map_fn)
                else:
                    raise NotImplementedError(
                        "Only Function expressions are supported for now."
                    )
            else:
                if isinstance(expr_or_fun_data, FunctionExpression):
                    output = self.parse_func_expr(expr_or_fun, map_fn=map_fn)
                    return output
                else:
                    raise NotImplementedError(
                        f"Only function call expressions are supported for now. Got {expr_or_fun_data}"
                    )
        else:
            raise ValueError(f"expr_or_fun should be a Parameter. Got {expr_or_fun}")

    def execute_func(
        self, func: Union[Function, Parameter], map_fn: Callable = lambda x: x.data
    ) -> Union[FunctionOutput, Parameter]:
        r"""Execute the function. If the function is async, use asyncio.run to execute it."""

        if isinstance(func, Parameter):
            try:

                call_func_tool = CallFunctionTool()
                func.add_successor_map_fn(call_func_tool, map_fn=map_fn)
                return call_func_tool.forward(func, context=self.context)

            except Exception as e:
                log.error(f"Error {e} executing function: {func.data}")
                error_msg = f"Error {e} executing function: {func.data}"
                return error_msg

        else:
            try:
                tool: FunctionTool = self.context[func.name]
                if tool.is_async:
                    return run_async_in_new_loop(tool.acall(*func.args, **func.kwargs))

                else:
                    output = tool.call(*func.args, **func.kwargs)
                    return output
            except Exception as e:
                log.error(f"Error {e} executing function: {func}")
                raise ValueError(f"Error {e} executing function: {func}")

    async def execute_func_async(self, func: Function) -> FunctionOutput:
        r"""Execute the function. If the function is sync, use await to execute it."""
        try:
            tool: FunctionTool = self.context[func.name]
            if tool.is_async:
                return await tool.acall(*func.args, **func.kwargs)
            else:
                return asyncio.to_thread(self.call, *func.args, **func.kwargs)
        except Exception as e:
            log.error(f"Error {e} executing function: {func}")
            raise ValueError(f"Error {e} executing function: {func}")

    def execute_func_expr(
        self,
        expr: Union[FunctionExpression, Parameter],
        map_fn: Callable = lambda x: x.data,
    ) -> Union[FunctionOutput, Parameter]:
        r"""Execute the function expression. Support both sync and async functions."""

        if isinstance(expr, Parameter):

            func: Parameter = self.parse_func_expr(expr, map_fn=map_fn)
            if not isinstance(func, Parameter):
                raise ValueError(f"Error parsing function expression: {expr}")

            # execute the function
            output: Parameter = self.execute_func(func)
            if not isinstance(output, Parameter):
                raise ValueError(f"Error executing function expression: {expr}")
            output.predecessors.add(expr)
            return output
        else:

            try:
                func: Function = self.parse_func_expr(expr)
                if not isinstance(func, Function):
                    raise ValueError(f"Error parsing function expression: {expr}")

                return self.execute_func(func)
            except Exception as e:
                # NOTE: if the function expression is not a function call, try to execute it as a function expression
                log.error(f"Error {e} executing function expression: {expr}")
                return FunctionOutput(
                    name=expr.action, input=expr, output=None, error=None
                )

    async def execute_func_expr_async(self, expr: FunctionExpression) -> FunctionOutput:
        r"""Execute the function expression. Support both sync and async functions."""
        func: Function = self.parse_func_expr(expr)
        try:
            return await self.execute_func_async(func)
        except Exception as e:
            # NOTE: if the function expression is not a function call, try to execute it as a function expression
            log.error(f"Error {e} executing function expression: {expr}")
            raise ValueError(f"Error {e} executing function expression: {expr}")

    def execute_func_expr_via_sandbox(self, expr: FunctionExpression) -> FunctionOutput:
        r"""Execute the function expression via sandbox. Only support sync functions."""
        func_output = FunctionOutput(
            name=expr.action, input=expr, output=None, error=None
        )
        try:
            action = (
                "output = " + expr.action
                if not expr.action.startswith("output")
                else expr.action
            )
            result = sandbox_exec(action, self.context)
            output = result.get("output", None)
            error = result.get("error", None)
            func_output.output = output
            func_output.error = error

        except Exception as e:
            log.error(f"Error {e} executing function expression: {expr}")
            raise ValueError(f"Error {e} executing function expression: {expr}")

        return func_output

    def execute_func_expr_via_eval(self, expr: FunctionExpression) -> FunctionOutput:
        r"""Execute the function expression via eval. Only support sync functions."""
        try:
            result = eval(expr.action, self.context)
            return FunctionOutput(
                name=expr.action,
                input=expr,
                output=result,
                error=None,
            )
        except Exception as e:
            log.error(f"Error {e} executing function expression: {expr}")
            raise ValueError(f"Error {e} executing function expression: {expr}")

    def _extra_repr(self) -> str:
        s = f"Tools: {self.tools}, Additional Context: {self._additional_context}"
        return s


if __name__ == "__main__":
    # test tool manager
    from adalflow.core.func_tool import FunctionTool
    from adalflow.components.model_client import OpenAIClient
    from adalflow.core.generator import Generator
    from adalflow.optim.parameter import Parameter
    from adalflow.utils import setup_env, printc

    setup_env()

    llm = Generator(
        model_client=OpenAIClient(),
        model_kwargs={"model": "gpt-3.5-turbo"},
    )
    # llm.train()

    def llm_as_tool(input: str, id: Optional[str] = None) -> str:
        """Used as a calculator tool."""
        printc(f"llm_as_tool: {input}", color="yellow")

        return llm(prompt_kwargs={"input_str": input}, id=id)

    llm_tool = FunctionTool(llm_as_tool, component=llm)
    # llm_tool.train()
    # output: Parameter = llm_tool("What is 2+2?")
    # output.draw_graph()
    # print(output)

    tool_manager = ToolManager(tools=[llm_tool])
    tool_manager.train()
    expr_or_fun = Parameter(
        name="expr_or_fun",
        data=FunctionExpression(action="llm_as_tool('What is 2+2?')"),
        eval_input="What is 2+2?",
        param_type=ParameterType.INPUT,
    )
    output: Parameter = tool_manager(expr_or_fun=expr_or_fun, step="parse")
    print(output)
    print(output.predecessors)
    assert len(output.predecessors) == 1
    # output = tool_manager(output, step="execute")
    # print(output)
    # output.draw_graph()

    # expr_or_fun = FunctionExpression(action="llm_as_tool('What is 2+2?')")

    # tool_manager.eval()
    # output = tool_manager(expr_or_fun=expr_or_fun, step="execute")
    # print(output)



"""Groq ModelClient integration."""

import os
from typing import Dict, Sequence, Optional, Any, TypeVar
import backoff
import logging
from adalflow.core.model_client import ModelClient
from adalflow.core.types import ModelType, CompletionUsage, GeneratorOutput


from adalflow.utils.lazy_import import safe_import, OptionalPackages

# optional import
groq = safe_import(OptionalPackages.GROQ.value[0], OptionalPackages.GROQ.value[1])


from groq import Groq, AsyncGroq
from groq import (
    APITimeoutError,
    InternalServerError,
    RateLimitError,
    UnprocessableEntityError,
)
from groq.types.chat import ChatCompletion as GroqChatCompletion
from groq.types import CompletionUsage as GroqCompletionUsage

T = TypeVar("T")
log = logging.getLogger(__name__)


class GroqAPIClient(ModelClient):
    __doc__ = r"""A component wrapper for the Groq API client.

    Visit https://console.groq.com/docs/ for more api details.
    Check https://console.groq.com/docs/models for the available models.

    Tested Groq models: 4/22/2024
    - llama3-8b-8192
    - llama3-70b-8192
    - mixtral-8x7b-32768
    - gemma-7b-it
    """

    def __init__(self, api_key: Optional[str] = None):
        r"""It is recommended to set the GROQ_API_KEY environment variable instead of passing it as an argument.

        Args:
            api_key (Optional[str], optional): Groq API key. Defaults to None.
        """
        super().__init__()
        self._api_key = api_key

        self.sync_client = self.init_sync_client()
        self.async_client = None  # only initialize if the async call is called

    def init_sync_client(self):
        api_key = self._api_key or os.getenv("GROQ_API_KEY")
        if not api_key:
            raise ValueError("Environment variable GROQ_API_KEY must be set")
        return Groq(api_key=api_key)

    def init_async_client(self):
        api_key = self._api_key or os.getenv("GROQ_API_KEY")
        if not api_key:
            raise ValueError("Environment variable GROQ_API_KEY must be set")
        return AsyncGroq(api_key=api_key)

    def parse_chat_completion(
        self, completion: "GroqChatCompletion"
    ) -> "GeneratorOutput":
        """
        Parse the completion to a string output.
        """
        log.debug(f"completion: {completion}")
        try:
            data = completion.choices[0].message.content
            usage = self.track_completion_usage(completion)
            return GeneratorOutput(data=None, usage=usage, raw_response=data)
        except Exception as e:
            log.error(f"Error parsing completion: {e}")
            return GeneratorOutput(data=str(e), usage=None, raw_response=completion)

    def track_completion_usage(self, completion: Any) -> CompletionUsage:
        usage: GroqCompletionUsage = completion.usage
        return CompletionUsage(
            completion_tokens=usage.completion_tokens,
            prompt_tokens=usage.prompt_tokens,
            total_tokens=usage.total_tokens,
        )

    def convert_inputs_to_api_kwargs(
        self,
        input: Optional[Any] = None,
        model_kwargs: Dict = {},
        model_type: ModelType = ModelType.UNDEFINED,
    ) -> Dict:
        final_model_kwargs = model_kwargs.copy()
        if model_type == ModelType.LLM:
            messages: Sequence[Dict[str, str]] = []
            if input is not None and input != "":
                messages.append({"role": "system", "content": input})
            final_model_kwargs["messages"] = messages
        else:
            raise ValueError(f"model_type {model_type} is not supported")
        return final_model_kwargs

    @backoff.on_exception(
        backoff.expo,
        (
            APITimeoutError,
            InternalServerError,
            RateLimitError,
            UnprocessableEntityError,
        ),
        max_time=5,
    )
    def call(self, api_kwargs: Dict = {}, model_type: ModelType = ModelType.UNDEFINED):
        assert (
            "model" in api_kwargs
        ), f"model must be specified in api_kwargs: {api_kwargs}"
        if model_type == ModelType.LLM:
            completion = self.sync_client.chat.completions.create(**api_kwargs)
            return completion
        else:
            raise ValueError(f"model_type {model_type} is not supported")

    @backoff.on_exception(
        backoff.expo,
        (
            APITimeoutError,
            InternalServerError,
            RateLimitError,
            UnprocessableEntityError,
        ),
        max_time=5,
    )
    async def acall(
        self, api_kwargs: Dict = {}, model_type: ModelType = ModelType.UNDEFINED
    ):
        if self.async_client is None:
            self.async_client = self.init_async_client()
        assert "model" in api_kwargs, "model must be specified"
        if model_type == ModelType.LLM:
            completion = await self.async_client.chat.completions.create(**api_kwargs)
            return completion
        else:
            raise ValueError(f"model_type {model_type} is not supported")

    @classmethod
    def from_dict(cls: type[T], data: Dict[str, Any]) -> T:
        obj = super().from_dict(data)
        # recreate the existing clients
        obj.sync_client = obj.init_sync_client()
        obj.async_client = obj.init_async_client()
        return obj

    def to_dict(self) -> Dict[str, Any]:
        r"""Convert the component to a dictionary."""
        # TODO: not exclude but save yes or no for recreating the clients
        exclude = [
            "sync_client",
            "async_client",
        ]  # unserializable object
        output = super().to_dict(exclude=exclude)
        return output



"""Leverage a Qdrant collection to retrieve documents."""

from typing import List, Optional, Any
from qdrant_client import QdrantClient, models

from adalflow.core.retriever import (
    Retriever,
)
from adalflow.core.embedder import Embedder

from adalflow.core.types import (
    RetrieverOutput,
    RetrieverStrQueryType,
    RetrieverStrQueriesType,
    Document,
)


class QdrantRetriever(Retriever[Any, RetrieverStrQueryType]):
    __doc__ = r"""Use a Qdrant collection to retrieve documents.

    Args:
        collection_name (str): the collection name in Qdrant.
        client (QdrantClient): An instance of qdrant_client.QdrantClient.
        embedder (Embedder): An instance of Embedder.
        top_k (Optional[int], optional): top k documents to fetch. Defaults to 10.
        vector_name (Optional[str], optional): the name of the vector in the collection. Defaults to None.
        text_key (str, optional): the key in the payload that contains the text. Defaults to "text".
        metadata_key (str, optional): the key in the payload that contains the metadata. Defaults to "meta_data".
        filter (Optional[models.Filter], optional): the filter to apply to the query. Defaults to None.

    References:
    [1] Qdrant: https://qdrant.tech/
    [2] Documentation: https://qdrant.tech/documentation/
    """

    def __init__(
        self,
        collection_name: str,
        client: QdrantClient,
        embedder: Embedder,
        top_k: Optional[int] = 10,
        vector_name: Optional[str] = None,
        text_key: str = "text",
        metadata_key: str = "meta_data",
        filter: Optional[models.Filter] = None,
    ):
        super().__init__()
        self._top_k = top_k
        self._collection_name = collection_name
        self._client = client
        self._embedder = embedder
        self._text_key = text_key
        self._metadata_key = metadata_key
        self._filter = filter

        self._vector_name = vector_name or self._get_first_vector_name()

    def reset_index(self):
        if self._client.collection_exists(self._collection_name):
            self._client.delete_collection(self._collection_name)

    def call(
        self,
        input: RetrieverStrQueriesType,
        top_k: Optional[int] = None,
        **kwargs,
    ) -> List[RetrieverOutput]:
        top_k = top_k or self._top_k
        queries: List[str] = input if isinstance(input, list) else [input]

        queries_embeddings = self._embedder(queries)

        query_requests: List[models.QueryRequest] = []
        for idx, query in enumerate(queries):
            query_embedding = queries_embeddings.data[idx].embedding
            query_requests.append(
                models.QueryRequest(
                    query=query_embedding,
                    limit=top_k,
                    using=self._vector_name,
                    with_payload=True,
                    with_vector=True,
                    filter=self._filter,
                    **kwargs,
                )
            )

        results = self._client.query_batch_points(
            self._collection_name, requests=query_requests
        )
        retrieved_outputs: List[RetrieverOutput] = []
        for result in results:
            out = self._points_to_output(
                result.points,
                query,
                self._text_key,
                self._metadata_key,
                self._vector_name,
            )
            retrieved_outputs.append(out)

        return retrieved_outputs

    def _get_first_vector_name(self) -> Optional[str]:
        vectors = self._client.get_collection(
            self._collection_name
        ).config.params.vectors

        if not isinstance(vectors, dict):
            # The collection only has the default, unnamed vector
            return None

        first_vector_name = list(vectors.keys())[0]

        # The collection has multiple vectors. Could also include the falsy unnamed vector - Empty string("")
        return first_vector_name or None

    @classmethod
    def _points_to_output(
        cls,
        points: List[models.ScoredPoint],
        query: str,
        text_key: str,
        metadata_key: str,
        vector_name: Optional[str],
    ) -> RetrieverOutput:
        doc_indices = [point.id for point in points]
        doc_scores = [point.score for point in points]
        documents = [
            cls._doc_from_point(point, text_key, metadata_key, vector_name)
            for point in points
        ]
        return RetrieverOutput(
            doc_indices=doc_indices,
            doc_scores=doc_scores,
            query=query,
            documents=documents,
        )

    @classmethod
    def _doc_from_point(
        cls,
        point: models.ScoredPoint,
        text_key: str,
        metadata_key: str,
        vector_name: Optional[str] = None,
    ) -> Document:
        vector = point.vector
        if isinstance(vector, dict):
            vector = vector[vector_name]

        payload = point.payload.copy()
        return Document(
            id=point.id,
            text=payload.get(text_key, ""),
            meta_data=payload.get(metadata_key, {}),
            vector=vector,
        )



"""Class prompt builder for AdalFlow system prompt."""

from typing import Dict, Any, Optional, List, TypeVar, Union
import logging
from functools import lru_cache

from jinja2 import Template, Environment, StrictUndefined, meta


from adalflow.core.default_prompt_template import DEFAULT_ADALFLOW_SYSTEM_PROMPT
from adalflow.optim.parameter import Parameter
from adalflow.core.component import DataComponent


logger = logging.getLogger(__name__)

T = TypeVar("T")


class Prompt(DataComponent):
    __doc__ = r"""Renders a text string(prompt) from a Jinja2 template string.

    In default, we use the :ref:`DEFAULT_ADALFLOW_SYSTEM_PROMPT<core-default_prompt_template>`  as the template.

    Args:
        template (str, optional): The Jinja2 template string. Defaults to DEFAULT_ADALFLOW_SYSTEM_PROMPT.
        preset_prompt_kwargs (Optional[Dict], optional): The preset prompt kwargs to fill in the variables in the prompt. Defaults to {}.

    Examples:
        >>> from core.prompt_builder import Prompt
        >>> prompt = Prompt(prompt_kwargs={"task_desc_str": "You are a helpful assistant."})
        >>> print(prompt)
        >>> prompt.print_prompt_template()
        >>> prompt.print_prompt(context_str="This is a context string.")
        >>> prompt.call(context_str="This is a context string.")

        When examples_str itself is another template with variables, You can use another Prompt to render it.

        >>> EXAMPLES_TEMPLATE = r'''
        >>> {% if examples %}
        >>> {% for example in examples %}
        >>> {{loop.index}}. {{example}}
        >>> {% endfor %}
        >>> {% endif %}
        >>> '''
        >>> examples_prompt = Prompt(template=EXAMPLES_TEMPLATE)
        >>> examples_str = examples_prompt.call(examples=["Example 1", "Example 2"])
        >>> # pass it to the main prompt
        >>> prompt.print_prompt(examples_str=examples_str)
    """
    # # save these two fields for serialization, using to_dict and from_dict
    # template: str = field(
    #     default=DEFAULT_ADALFLOW_SYSTEM_PROMPT,
    #     metadata={"desc": "The Jinja2 template string."},
    # )
    # prompt_kwargs: Dict[str, Parameter] = field(
    #     default_factory=dict, metadata={"desc": "The preset prompt kwargs."}
    # )
    # prompt_va

    def __init__(
        self,
        template: Optional[str] = None,
        prompt_kwargs: Optional[Dict[str, Union[Any, Parameter]]] = {},
    ):
        super().__init__()

        self.template = template or DEFAULT_ADALFLOW_SYSTEM_PROMPT
        self.__create_jinja2_template()
        self.prompt_variables: List[str] = []
        for var in self._find_template_variables(self.template):
            self.prompt_variables.append(var)

        logger.info(f"{__class__.__name__} has variables: {self.prompt_variables}")

        self.prompt_kwargs = prompt_kwargs

    def __create_jinja2_template(self):
        r"""Create the Jinja2 template object."""
        try:
            self.jinja2_template: Template = get_jinja2_environment().from_string(
                self.template
            )
        except Exception as e:
            raise ValueError(f"Invalid Jinja2 template: {e}")

    def update_prompt_kwargs(self, **kwargs):
        r"""Update the initial prompt kwargs after Prompt is initialized."""
        self.prompt_kwargs.update(kwargs)

    def get_prompt_variables(self) -> List[str]:
        r"""Get the prompt kwargs."""
        return self.prompt_variables

    def is_key_in_template(self, key: str) -> bool:
        r"""Check if the key exists in the template."""
        return key in self.prompt_variables

    def _find_template_variables(self, template_str: str):
        """Automatically find all the variables in the template."""
        parsed_content = self.jinja2_template.environment.parse(template_str)
        return meta.find_undeclared_variables(parsed_content)

    def compose_prompt_kwargs(self, **kwargs) -> Dict:
        r"""Compose the final prompt kwargs by combining the initial and the provided kwargs at runtime."""
        composed_kwargs = {key: None for key in self.prompt_variables}
        if self.prompt_kwargs:
            composed_kwargs.update(self.prompt_kwargs)
        if kwargs:
            for key, _ in kwargs.items():
                if key not in composed_kwargs:
                    logger.debug(f"Key {key} does not exist in the prompt_kwargs.")
            composed_kwargs.update(kwargs)
        return composed_kwargs

    def print_prompt_template(self):
        r"""Print the template string."""
        print("Template:")
        print("-------")
        print(f"{self.template}")
        print("-------")

    def print_prompt(self, **kwargs) -> str:
        r"""Print the rendered prompt string using the preset_prompt_kwargs and the provided kwargs."""
        try:
            pass_kwargs = self.compose_prompt_kwargs(**kwargs)
            pass_kwargs = _convert_prompt_kwargs_to_str(pass_kwargs)
            logger.debug(f"Prompt kwargs: {pass_kwargs}")

            prompt_str = self.jinja2_template.render(**pass_kwargs)
            print("Prompt:\n______________________")
            print(prompt_str)
            return prompt_str
        except Exception as e:
            raise ValueError(f"Error rendering Jinja2 template: {e}")

    # def __call__(self, *args: Any, **kwds: Any) -> Any:
    #     return self.call(*args, **kwds)

    def call(self, **kwargs) -> str:
        """
        Renders the prompt template with keyword arguments. Allow None values.
        """
        try:
            pass_kwargs = self.compose_prompt_kwargs(**kwargs)
            pass_kwargs = _convert_prompt_kwargs_to_str(pass_kwargs)
            prompt_str = self.jinja2_template.render(**pass_kwargs)
            return prompt_str

        except Exception as e:
            raise ValueError(f"Error rendering Jinja2 template: {e}")

    def _extra_repr(self) -> str:
        s = f"template: {self.template}"
        prompt_kwargs_str = _convert_prompt_kwargs_to_str(self.prompt_kwargs)
        if prompt_kwargs_str:
            s += f", prompt_kwargs: {prompt_kwargs_str}"
        if self.prompt_variables:
            s += f", prompt_variables: {self.prompt_variables}"
        return s

    def __repr__(self) -> str:
        s = f"template: {self.template}"
        prompt_kwargs_str = _convert_prompt_kwargs_to_str(self.prompt_kwargs)
        if prompt_kwargs_str:
            s += f", prompt_kwargs: {prompt_kwargs_str}"
        if self.prompt_variables:
            s += f", prompt_variables: {self.prompt_variables}"
        return s

    @classmethod
    def from_dict(cls: type[T], data: Dict[str, Any]) -> T:
        obj = super().from_dict(data)
        # recreate the jinja2 template
        obj.jinja2_template = get_jinja2_environment().from_string(obj.template)
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Get the dictionary representation of all the Prompt object's attributes, with sorting applied to
        dictionary keys and list elements to ensure consistent ordering.
        """
        exclude = ["jinja2_template"]  # unserializable object
        output = super().to_dict(exclude=exclude)
        return output


def _convert_prompt_kwargs_to_str(prompt_kwargs: Dict) -> Dict[str, str]:
    r"""Convert the prompt_kwargs to a dictionary with string values."""
    prompt_kwargs_str: Dict[str, str] = {}

    for key, p in prompt_kwargs.items():

        if isinstance(p, Parameter):

            prompt_kwargs_str[key] = p.data
        elif isinstance(p, list):
            prompt_kwargs_str[key] = [
                (
                    p_elem.data_in_prompt(p_elem)
                    if isinstance(p_elem, Parameter)
                    else p_elem
                )
                for p_elem in p
            ]

        else:
            prompt_kwargs_str[key] = p
    return prompt_kwargs_str


@lru_cache(None)
def get_jinja2_environment():
    r"""Helper function for Prompt component to get the Jinja2 environment with the default settings."""
    try:
        default_environment = Environment(
            undefined=StrictUndefined,
            trim_blocks=True,
            keep_trailing_newline=True,
            lstrip_blocks=True,
        )

        return default_environment
    except Exception as e:
        raise ValueError(f"Invalid Jinja2 environment: {e}")



"""Semantic search/embedding-based retriever using FAISS."""

import faiss
from typing import (
    List,
    Optional,
    Sequence,
    Union,
    Dict,
    overload,
    Literal,
    Any,
    Callable,
)
import numpy as np
import logging
import os


from adalflow.core.retriever import Retriever
from adalflow.core.embedder import Embedder
from adalflow.core.types import (
    RetrieverOutput,
    RetrieverOutputType,
    RetrieverStrQueryType,
    EmbedderOutputType,
)
from adalflow.core.functional import normalize_np_array, is_normalized

from adalflow.utils.lazy_import import safe_import, OptionalPackages

safe_import(OptionalPackages.FAISS.value[0], OptionalPackages.FAISS.value[1])

log = logging.getLogger(__name__)

# single embedding
FAISSRetrieverDocumentEmbeddingType = Union[List[float], np.ndarray]
FAISSRetrieverDocumentsType = Sequence[FAISSRetrieverDocumentEmbeddingType]

FAISSRetrieverEmbeddingQueryType = Union[
    List[float], List[List[float]], np.ndarray
]  # single embedding or list of embeddings
FAISSRetrieverQueryType = Union[RetrieverStrQueryType, FAISSRetrieverEmbeddingQueryType]
FAISSRetrieverQueriesType = Sequence[FAISSRetrieverQueryType]
FAISSRetrieverQueriesStrType = Sequence[RetrieverStrQueryType]
FAISSRetrieverQueriesEmbeddingType = Sequence[FAISSRetrieverEmbeddingQueryType]

os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"


class FAISSRetriever(
    Retriever[FAISSRetrieverDocumentEmbeddingType, FAISSRetrieverQueryType]
):
    __doc__ = r"""Semantic search/embedding-based retriever using FAISS.

    To use the retriever, you can either pass the index embeddings from the :meth:`__init__` or use the :meth:`build_index_from_documents` method.


    Args:
        embedder (Embedder, optimal): The embedder component to use for converting the queries in string format to embeddings.
            Ensure the vectorizer is exactly the same as the one used to the embeddings in the index.
        top_k (int, optional): Number of chunks to retrieve. Defaults to 5.
        dimensions (Optional[int], optional): Dimension of the embeddings. Defaults to None. It can automatically infer the dimensions from the first chunk.
        documents (Optional[FAISSRetrieverDocumentType], optional): List of embeddings. Format can be List[List[float]] or List[np.ndarray]. Defaults to None.
        metric (Literal["cosine", "euclidean", "prob"], optional): The metric to use for the retrieval. Defaults to "prob" which converts cosine similarity to probability.

    How FAISS works:

    The retriever uses in-memory Faiss index to retrieve the top k chunks
    d: dimension of the vectors
    xb: number of vectors to put in the index
    xq: number of queries
    The data type dtype must be float32.

    Note: When the num of chunks are less than top_k, the last columns will be -1

    Other index options:
    - faiss.IndexFlatL2: L2 or Euclidean distance, [-inf, inf]
    - faiss.IndexFlatIP: Inner product of embeddings (inner product of normalized vectors will be cosine similarity, [-1, 1])

    We choose cosine similarity and convert it to range [0, 1] by adding 1 and dividing by 2 to simulate probability in [0, 1]

    Install FAISS:

    As FAISS is optional package, you can install it with pip for cpu version:
    ```bash
    pip install faiss-cpu
    ```
    For GPU version:
    You might have to use conda to install faiss-gpu:https://github.com/facebookresearch/faiss/wiki/Installing-Faiss

    References:
    - FAISS: https://github.com/facebookresearch/faiss
    """

    def __init__(
        self,
        embedder: Optional[Embedder] = None,
        top_k: int = 5,
        dimensions: Optional[int] = None,
        documents: Optional[Any] = None,
        document_map_func: Optional[
            Callable[[Any], FAISSRetrieverDocumentEmbeddingType]
        ] = None,
        metric: Literal["cosine", "euclidean", "prob"] = "prob",
    ):
        super().__init__()

        self.reset_index()

        self.dimensions = dimensions
        self.embedder = embedder  # used to vectorize the queries
        self.top_k = top_k
        self.metric = metric
        if self.metric == "cosine" or self.metric == "prob":
            self._faiss_index_type = faiss.IndexFlatIP
            self._needs_normalized_embeddings = True
        elif self.metric == "euclidean":
            self._faiss_index_type = faiss.IndexFlatL2
            self._needs_normalized_embeddings = False
        else:
            raise ValueError(f"Invalid metric: {self.metric}")

        if documents:
            self.documents = documents
            self.build_index_from_documents(documents, document_map_func)

    def reset_index(self):
        self.index = None
        self.total_documents: int = 0
        self.documents: Sequence[Any] = None
        self.xb: np.ndarray = None
        self.dimensions: Optional[int] = None
        self.indexed: bool = False

    def _preprare_faiss_index_from_np_array(self, xb: np.ndarray):
        r"""Prepare the faiss index from the numpy array."""
        if not self.dimensions:
            self.dimensions = self.xb.shape[1]
        else:
            assert (
                self.dimensions == self.xb.shape[1]
            ), f"Dimension mismatch: {self.dimensions} != {self.xb.shape[1]}"
        self.total_documents = xb.shape[0]

        self.index = self._faiss_index_type(self.dimensions)
        self.index.add(xb)
        self.indexed = True

    def build_index_from_documents(
        self,
        documents: Sequence[Any],
        document_map_func: Optional[
            Callable[[Any], FAISSRetrieverDocumentEmbeddingType]
        ] = None,
    ):
        r"""Build index from embeddings.

        Args:
            documents: List of embeddings. Format can be List[List[float]] or List[np.ndarray]

        If you are using Document format, pass them as [doc.vector for doc in documents]
        """
        if document_map_func:
            assert callable(document_map_func), "document_map_func should be callable"
            documents = [document_map_func(doc) for doc in documents]
        try:
            self.documents = documents

            # convert to numpy array
            if not isinstance(documents, np.ndarray) and isinstance(
                documents[0], Sequence
            ):
                # ensure all the embeddings are of the same size
                assert all(
                    len(doc) == len(documents[0]) for doc in documents
                ), "All embeddings should be of the same size"
                self.xb = np.array(documents, dtype=np.float32)
            else:
                self.xb = documents
            if self._needs_normalized_embeddings:
                first_vector = self.xb[0]
                if not is_normalized(first_vector):
                    log.warning(
                        "Embeddings are not normalized, normalizing the embeddings"
                    )
                    self.xb = normalize_np_array(self.xb)

            self._preprare_faiss_index_from_np_array(self.xb)
            log.info(f"Index built with {self.total_documents} chunks")
        except Exception as e:
            log.error(f"Error building index: {e}, resetting the index")
            # reset the index
            self.reset_index()
            raise e

    def _convert_cosine_similarity_to_probability(self, D: np.ndarray) -> np.ndarray:
        D = np.clip(D, -1, 1)
        D = (D + 1) / 2
        D = np.round(D, 3)
        return D

    def _to_retriever_output(
        self, Ind: np.ndarray, D: np.ndarray
    ) -> RetrieverOutputType:
        r"""Convert the indices and distances to RetrieverOutputType format."""
        output: RetrieverOutputType = []
        # Step 1: Filter out the -1, -1 columns along with its scores when top_k > len(chunks)
        if -1 in Ind:
            valid_columns = ~np.any(Ind == -1, axis=0)

            D = D[:, valid_columns]
            Ind = Ind[:, valid_columns]
        # Step 2: processing rows (one query at a time)
        for row in zip(Ind, D):
            indices, distances = row
            # convert from numpy to list
            retrieved_documents_indices = indices.tolist()
            retrieved_documents_scores = distances.tolist()
            output.append(
                RetrieverOutput(
                    doc_indices=retrieved_documents_indices,
                    doc_scores=retrieved_documents_scores,
                )
            )

        return output

    def retrieve_embedding_queries(
        self,
        input: FAISSRetrieverQueriesEmbeddingType,
        top_k: Optional[int] = None,
    ) -> RetrieverOutputType:
        if not self.indexed or self.index.ntotal == 0:
            raise ValueError(
                "Index is empty. Please set the chunks to build the index from"
            )
        # check if the input is List, convert to numpy array
        try:
            if not isinstance(input, np.ndarray):
                xq = np.array(input, dtype=np.float32)
            else:
                xq = input
        except Exception as e:
            log.error(f"Error converting input to numpy array: {e}")
            raise e

        D, Ind = self.index.search(xq, top_k if top_k else self.top_k)
        if self.metric == "prob":
            D = self._convert_cosine_similarity_to_probability(D)
        output: RetrieverOutputType = self._to_retriever_output(Ind, D)
        return output

    def retrieve_string_queries(
        self,
        input: Union[str, List[str]],
        top_k: Optional[int] = None,
    ) -> RetrieverOutputType:
        r"""Retrieve the top k chunks given the query or queries in string format.

        Args:
            input: The query or list of queries in string format. Note: ensure the maximum number of queries fits into the embedder.
            top_k: The number of chunks to retrieve. When top_k is not provided, it will use the default top_k set during initialization.

        When top_k is not provided, it will use the default top_k set during initialization.
        """
        if not self.indexed or self.index.ntotal == 0:
            raise ValueError(
                "Index is empty. Please set the chunks to build the index from"
            )
        queries = [input] if isinstance(input, str) else input
        # filter out empty queries
        valid_queries: List[str] = []
        record_map: Dict[int, int] = (
            {}
        )  # final index : the position in the initial queries
        for i, q in enumerate(queries):
            if not q:
                log.warning("Empty query found, skipping")
                continue
            valid_queries.append(q)
            record_map[len(valid_queries) - 1] = i
        # embed the queries, assume the length fits into a batch.
        try:
            embeddings: EmbedderOutputType = self.embedder(valid_queries)
            queries_embeddings: List[List[float]] = [
                data.embedding for data in embeddings.data
            ]

        except Exception as e:
            log.error(f"Error embedding queries: {e}")
            raise e
        xq = np.array(queries_embeddings, dtype=np.float32)
        D, Ind = self.index.search(xq, top_k if top_k else self.top_k)
        D = self._convert_cosine_similarity_to_probability(D)

        output: RetrieverOutputType = [
            RetrieverOutput(doc_indices=[], query=query) for query in queries
        ]
        retrieved_output: RetrieverOutputType = self._to_retriever_output(Ind, D)

        # fill in the doc_indices and score for valid queries
        for i, per_query_output in enumerate(retrieved_output):
            initial_index = record_map[i]
            output[initial_index].doc_indices = per_query_output.doc_indices
            output[initial_index].doc_scores = per_query_output.doc_scores

        return output

    @overload
    def call(
        self,
        input: FAISSRetrieverQueriesEmbeddingType,
        top_k: Optional[int] = None,
    ) -> RetrieverOutputType:
        r"""Retrieve the top k chunks given the query or queries in embedding format."""
        ...

    @overload
    def call(
        self,
        input: FAISSRetrieverQueriesStrType,
        top_k: Optional[int] = None,
    ) -> RetrieverOutputType:
        r"""Retrieve the top k chunks given the query or queries in string format."""
        ...

    def call(
        self,
        input: FAISSRetrieverQueriesType,
        top_k: Optional[int] = None,
    ) -> RetrieverOutputType:
        r"""Retrieve the top k chunks given the query or queries in embedding or string format."""
        assert (
            self.indexed
        ), "Index is not built. Please build the index using build_index_from_documents"
        if isinstance(input, str) or (
            isinstance(input, Sequence) and isinstance(input[0], str)
        ):
            assert self.embedder, "Embedder is not provided"
            return self.retrieve_string_queries(input, top_k)
        else:
            return self.retrieve_embedding_queries(input, top_k)

    def _extra_repr(self) -> str:
        s = f"top_k={self.top_k}"
        if self.metric:
            s += f", metric={self.metric}"
        if self.dimensions:
            s += f", dimensions={self.dimensions}"
        if self.documents:
            s += f", total_documents={self.total_documents}"
        return s



"""Functional data classes to support functional components like Generator, Retriever, and Assistant."""

from enum import Enum, auto
from typing import (
    List,
    Dict,
    Any,
    Optional,
    Union,
    Generic,
    TypeVar,
    Sequence,
    Literal,
    Callable,
    Awaitable,
)
from collections import OrderedDict
from dataclasses import (
    dataclass,
    field,
    InitVar,
)
from uuid import UUID
from datetime import datetime
import uuid
import logging
import json

from adalflow.core.base_data_class import DataClass, required_field
from adalflow.core.tokenizer import Tokenizer
from adalflow.core.functional import (
    is_normalized,
    generate_function_call_expression_from_callable,
)
from adalflow.components.model_client import (
    CohereAPIClient,
    TransformersClient,
    AnthropicAPIClient,
    GroqAPIClient,
    OpenAIClient,
    GoogleGenAIClient,
    OllamaClient,
)


logger = logging.getLogger(__name__)

T_co = TypeVar("T_co", covariant=True)
T = TypeVar("T")  # invariant type


#######################################################################################
# Data modeling for ModelClient
######################################################################################
class ModelType(Enum):
    __doc__ = r"""The type of the model, including Embedder, LLM, Reranker.
     It helps ModelClient identify the model type required to correctly call the model."""
    EMBEDDER = auto()
    LLM = auto()
    RERANKER = auto()  # ranking model
    IMAGE_GENERATION = auto()  # image generation models like DALL-E
    UNDEFINED = auto()


@dataclass
class ModelClientType:
    __doc__ = r"""A quick way to access all model clients in the ModelClient module.

    From this:

    .. code-block:: python

        from adalflow.components.model_client import CohereAPIClient, TransformersClient, AnthropicAPIClient, GroqAPIClient, OpenAIClient

        model_client = OpenAIClient()

    To this:

    .. code-block:: python

        from adalflow.core.types import ModelClientType

        model_client = ModelClientType.OPENAI()
    """
    COHERE = CohereAPIClient
    TRANSFORMERS = TransformersClient
    ANTHROPIC = AnthropicAPIClient
    GROQ = GroqAPIClient
    OPENAI = OpenAIClient
    GOOGLE_GENAI = GoogleGenAIClient
    OLLAMA = OllamaClient


# TODO: define standard required outputs
def get_model_args(model_type: ModelType) -> List[str]:
    r"""Get the required keys in model_kwargs for a specific model type.

    note:
    If your model inference sdk uses different keys, you need to convert them to the standard keys here in their specifc ModelClient.

    Args:
        model_type (ModelType): The model type

    Returns:
        List[str]: The required keys in model_kwargs
    """
    if model_type == ModelType.EMBEDDER:
        return ["model"]
    elif model_type == ModelType.LLM:
        return ["model"]
    elif model_type == ModelType.RERANKER:
        return ["model", "top_k", "documents", "query"]
    else:
        return []


#######################################################################################
# Data modeling for Embedder component
######################################################################################
@dataclass
class Embedding:
    """
    Container for a single embedding.

    In sync with api spec, same as openai/types/embedding.py
    """

    embedding: List[float]
    index: Optional[int]  # match with the index of the input, in case some are missing


@dataclass
class Usage:
    """
    In sync with OpenAI embedding api spec, same as openai/types/create_embedding_response.py
    """

    prompt_tokens: int
    total_tokens: int


@dataclass
class EmbedderOutput(DataClass):
    __doc__ = r"""Container to hold the response from an Embedder model. Only Per-batch.

    Data standard for Embedder model output to interact with other components.
    Batch processing is often available, thus we need a list of Embedding objects.
    """

    data: List[Embedding] = field(
        default_factory=list, metadata={"desc": "List of embeddings"}
    )
    model: Optional[str] = field(default=None, metadata={"desc": "Model name"})
    usage: Optional[Usage] = field(default=None, metadata={"desc": "Usage tracking"})
    error: Optional[str] = field(default=None, metadata={"desc": "Error message"})
    raw_response: Optional[Any] = field(
        default=None, metadata={"desc": "Raw response"}
    )  # only used if error
    input: Optional[List[str]] = field(default=None, metadata={"desc": "Input text"})

    @property
    def length(self) -> int:
        return len(self.data) if self.data and isinstance(self.data, Sequence) else 0

    @property
    def embedding_dim(self) -> int:
        r"""The dimension of the embedding, assuming all embeddings have the same dimension.

        Returns:
            int: The dimension of the embedding, -1 if no embedding is available
        """
        return (
            len(self.data[0].embedding) if self.data and self.data[0].embedding else -1
        )

    @property
    def is_normalized(self) -> bool:
        r"""Check if the embeddings are normalized to unit vectors.

        Returns:
            bool: True if the embeddings are normalized, False otherwise
        """
        return (
            is_normalized(self.data[0].embedding)
            if self.data and self.data[0].embedding
            else False
        )


EmbedderInputType = Union[str, Sequence[str]]
EmbedderOutputType = EmbedderOutput

BatchEmbedderInputType = EmbedderInputType
BatchEmbedderOutputType = List[EmbedderOutputType]


#######################################################################################
# Data modeling for Generator component
######################################################################################
@dataclass
class TokenLogProb:
    r"""similar to openai.ChatCompletionTokenLogprob"""

    token: str
    logprob: float


@dataclass
class CompletionUsage:
    __doc__ = r"In sync with OpenAI completion usage api spec at openai/types/completion_usage.py"
    completion_tokens: Optional[int] = field(
        metadata={"desc": "Number of tokens in the generated completion"}, default=None
    )
    prompt_tokens: Optional[int] = field(
        metadata={"desc": "Number of tokens in the prompt"}, default=None
    )
    total_tokens: Optional[int] = field(
        metadata={
            "desc": "Total number of tokens used in the request (prompt + completion)"
        },
        default=None,
    )


@dataclass
class GeneratorOutput(DataClass, Generic[T_co]):
    __doc__ = r"""
    The output data class for the Generator component.
    We can not control its output 100%, so we use this to track the error_message and
    allow the raw string output to be passed through.

    (1) When model predict and output processors are both without error,
    we have data as the final output, error as None.
    (2) When either model predict or output processors have error,
    we have data as None, error as the error message.

    Raw_response will depends on the model predict.
    """
    id: Optional[str] = field(
        default=None, metadata={"desc": "The unique id of the output"}
    )

    data: T_co = field(
        default=None,
        metadata={"desc": "The final output data potentially after output parsers"},
    )
    error: Optional[str] = field(
        default=None,
        metadata={"desc": "Error message if any"},
    )
    usage: Optional[CompletionUsage] = field(
        default=None, metadata={"desc": "Usage tracking"}
    )
    raw_response: Optional[str] = field(
        default=None, metadata={"desc": "Raw string response from the model"}
    )  # parsed from model client response
    metadata: Optional[Dict[str, object]] = field(
        default=None, metadata={"desc": "Additional metadata"}
    )


GeneratorOutputType = GeneratorOutput[object]

#######################################################################################
# Data modeling for Retriever component
######################################################################################

RetrieverQueryType = TypeVar("RetrieverQueryType", contravariant=True)
RetrieverStrQueryType = str
RetrieverQueriesType = Union[RetrieverQueryType, Sequence[RetrieverQueryType]]
RetrieverStrQueriesType = Union[str, Sequence[RetrieverStrQueryType]]

RetrieverDocumentType = TypeVar("RetrieverDocumentType", contravariant=True)
RetrieverStrDocumentType = str  # for text retrieval
RetrieverDocumentsType = Sequence[RetrieverDocumentType]


@dataclass
class RetrieverOutput(DataClass):
    __doc__ = r"""Save the output of a single query in retrievers.

    It is up to the subclass of Retriever to specify the type of query and document.
    """

    id: str = field(default=None, metadata={"desc": "The unique id of the output"})

    doc_indices: List[int] = field(
        default=required_field, metadata={"desc": "List of document indices"}
    )
    doc_scores: List[float] = field(
        default=None, metadata={"desc": "List of document scores"}
    )
    query: RetrieverQueryType = field(
        default=None, metadata={"desc": "The query used to retrieve the documents"}
    )
    documents: List[RetrieverDocumentType] = field(
        default=None, metadata={"desc": "List of retrieved documents"}
    )


RetrieverOutputType = Union[
    List[RetrieverOutput], RetrieverOutput
]  # so to support multiple queries at once


#######################################################################################
# Data modeling for function calls
######################################################################################
AsyncCallable = Callable[..., Awaitable[Any]]


@dataclass
class FunctionDefinition(DataClass):
    __doc__ = r"""The data modeling of a function definition, including the name, description, and parameters."""
    class_instance: Optional[Any] = field(
        default=None,
        metadata={"desc": "The instance of the class this function belongs to"},
    )
    func_name: str = field(
        metadata={"desc": "The name of the tool"}, default=required_field
    )
    func_desc: Optional[str] = field(
        default=None, metadata={"desc": "The description of the tool"}
    )
    func_parameters: Dict[str, object] = field(
        default_factory=dict, metadata={"desc": "The schema of the parameters"}
    )

    def fn_schema_str(self, type: Literal["json", "yaml"] = "json") -> str:
        r"""Get the function definition str to be used in the prompt.

        You should also directly use :meth:`to_json` and :meth:`to_yaml` to get the schema in JSON or YAML format.
        """
        if type == "json":
            return self.to_json()
        elif type == "yaml":
            return self.to_yaml()
        else:
            raise ValueError(f"Unsupported type: {type}")


@dataclass
class Function(DataClass):
    __doc__ = r"""The data modeling of a function call, including the name and keyword arguments.

    You can use the exclude in :meth:`to_json` and :meth:`to_yaml` to exclude the `thought` field if you do not want to use chain-of-thought pattern.

    Example:

    .. code-block:: python

        # assume the function is added in a context_map
        # context_map = {"add": add}

        def add(a, b):
            return a + b

        # call function add with arguments 1 and 2
        fun = Function(name="add", kwargs={"a": 1, "b": 2})
        # evaluate the function
        result = context_map[fun.name](**fun.kwargs)

        # or call with positional arguments
        fun = Function(name="add", args=[1, 2])
        result = context_map[fun.name](*fun.args)
    """
    thought: Optional[str] = field(
        default=None, metadata={"desc": "Why the function is called"}
    )
    name: str = field(default="", metadata={"desc": "The name of the function"})
    args: Optional[List[object]] = field(
        default_factory=list,
        metadata={"desc": "The positional arguments of the function"},
    )
    kwargs: Optional[Dict[str, object]] = field(
        default_factory=dict,
        metadata={"desc": "The keyword arguments of the function"},
    )

    @classmethod
    def from_function(
        cls,
        func: Union[Callable[..., Any], AsyncCallable],
        thought: Optional[str] = None,
        *args,
        **kwargs,
    ) -> "Function":
        r"""Create a Function object from a function.

        Args:
            fun (Union[Callable[..., Any], AsyncCallable]): The function to be converted

        Returns:
            Function: The Function object

        Usage:
        1. Create a Function object from a function call:
        2. use :meth:`to_json` and :meth:`to_yaml` to get the schema in JSON or YAML format.
        3. This will be used as an example in prompt showing LLM how to call the function.

        Example:

        .. code-block:: python

            from adalflow.core.types import Function

            def add(a, b):
                return a + b

            # create a function call object with positional arguments
            fun = Function.from_function(add, thought="Add two numbers", 1, 2)
            print(fun)

            # output
            # Function(thought='Add two numbers', name='add', args=[1, 2])
        """
        return cls(
            thought=thought,
            name=func.__name__,
            args=args,
            kwargs=kwargs,
        )

    __output_fields__ = ["thought", "name", "kwargs"]


_action_desc = """FuncName(<kwargs>) \
Valid function call expression. \
Example: "FuncName(a=1, b=2)" \
Follow the data type specified in the function parameters.\
e.g. for Type object with x,y properties, use "ObjectType(x=1, y=2)"""


@dataclass
class FunctionExpression(DataClass):
    __doc__ = r"""The data modeling of a function expression for a call, including the name and arguments.

    Example:

    .. code-block:: python

        def add(a, b):
            return a + b

        # call function add with positional arguments 1 and 2
        fun_expr = FunctionExpression(action="add(1, 2)")
        # evaluate the expression
        result = eval(fun_expr.action)
        print(result)
        # Output: 3

        # call function add with keyword arguments
        fun_expr = FunctionExpression(action="add(a=1, b=2)")
        result = eval(fun_expr.action)
        print(result)
        # Output: 3

    Why asking LLM to generate function expression (code snippet) for a function call?
    - It is more efficient/compact to call a function.
    - It is more flexible.
        (1) for the full range of Python expressions, including arithmetic operations, nested function calls, and more.
        (2) allow to pass variables as arguments.
    - Ease of parsing using ``ast`` module.

    The benefits are less failed function calls.
    """
    # question: str = field(
    #     default=None, metadata={"desc": "The question to ask the LLM"}
    # )
    thought: str = field(default=None, metadata={"desc": "Why the function is called"})
    action: str = field(
        default_factory=required_field,
        metadata={"desc": _action_desc},
    )

    @classmethod
    def from_function(
        cls,
        func: Union[Callable[..., Any], AsyncCallable],
        thought: Optional[str] = None,
        *args,
        **kwargs,
    ) -> "FunctionExpression":
        r"""Create a FunctionExpression object from a function.

        Args:
            fun (Union[Callable[..., Any], AsyncCallable]): The function to be converted

        Returns:
            FunctionExpression: The FunctionExpression object

        Usage:
        1. Create a FunctionExpression object from a function call:
        2. use :meth:`to_json` and :meth:`to_yaml` to get the schema in JSON or YAML format.
        3. This will be used as an example in prompt showing LLM how to call the function.

        Example:

        .. code-block:: python

            from adalflow.core.types import FunctionExpression

            def add(a, b):
                return a + b

            # create an expression for the function call and using keyword arguments
            fun_expr = FunctionExpression.from_function(
                add, thought="Add two numbers", a=1, b=2
            )
            print(fun_expr)

            # output
            # FunctionExpression(thought='Add two numbers', action='add(a=1, b=2)')
        """
        try:
            action = generate_function_call_expression_from_callable(
                func, *args, **kwargs
            )
        except Exception as e:
            logger.error(f"Error generating function expression: {e}")
            raise ValueError(f"Error generating function expression: {e}")
        return cls(action=action, thought=thought)


@dataclass
class FunctionOutput(DataClass):
    __doc__ = (
        r"""The output of a tool, which could be a function, a class, or a module."""
    )
    name: Optional[str] = field(
        default=None, metadata={"desc": "The name of the function"}
    )
    input: Optional[Union[Function, FunctionExpression]] = field(
        default=None, metadata={"desc": "The Function or FunctionExpression object"}
    )
    parsed_input: Optional[Function] = field(
        default=None,
        metadata={
            "desc": "The parsed Function object if the input is FunctionExpression"
        },
    )
    output: Optional[object] = field(
        default=None, metadata={"desc": "The output of the function execution"}
    )
    error: Optional[str] = field(
        default=None, metadata={"desc": "The error message if any"}
    )


#######################################################################################
# Data modeling for agent component
######################################################################################
@dataclass
class StepOutput(DataClass, Generic[T]):
    __doc__ = r"""The output of a single step in the agent. Suits for serial planning agent such as React"""
    step: int = field(
        default=0, metadata={"desc": "The order of the step in the agent"}
    )

    # This action can be in Function, or Function Exptression, or just str
    # it includes the thought and action already
    # directly the output from planner LLMs
    action: T = field(
        default=None, metadata={"desc": "The action the agent takes at this step"}
    )

    function: Optional[Function] = field(
        default=None, metadata={"desc": "The parsed function from the action"}
    )

    observation: Optional[str] = field(
        default=None, metadata={"desc": "The execution result shown for this action"}
    )

    def to_prompt_str(self) -> str:
        output: Dict[str, Any] = {}
        if self.action and isinstance(self.action, FunctionExpression):
            if self.action.thought:
                output["thought"] = self.action.thought
            output["action"] = self.action.action if self.action else None
        if self.observation:
            output["observation"] = (
                self.observation.to_dict()
                if hasattr(self.observation, "to_dict")
                else str(self.observation)
            )
        return json.dumps(output)


#######################################################################################
# Data modeling for data processing pipleline such as Text splitting and Embedding
######################################################################################
@dataclass
class Document(DataClass):
    __doc__ = r"""A text container with optional metadata and vector representation.

    It is the data structure to support functions like Retriever, DocumentSplitter, and used with LocalDB.
    """

    text: str = field(metadata={"desc": "The main text"})

    meta_data: Optional[Dict[str, Any]] = field(
        default=None, metadata={"desc": "Metadata for the document"}
    )
    # can save data for filtering at retrieval time too
    vector: List[float] = field(
        default_factory=list,
        metadata={"desc": "The vector representation of the document"},
    )
    # the vector representation of the document

    id: Optional[str] = field(
        default_factory=lambda: str(uuid.uuid4()), metadata={"desc": "Unique id"}
    )  # unique id of the document
    order: Optional[int] = field(
        default=None,
        metadata={"desc": "Order of the chunked document in the original document"},
    )

    score: Optional[float] = field(
        default=None,
        metadata={"desc": "Score of the document, likely used in retrieval output"},
    )
    parent_doc_id: Optional[Union[str, UUID]] = field(
        default=None, metadata={"desc": "id of the Document where the chunk is from"}
    )

    estimated_num_tokens: Optional[int] = field(
        default=None,
        metadata={
            "desc": "Estimated number of tokens in the text, useful for cost estimation"
        },
    )

    def __post_init__(self):
        if self.estimated_num_tokens is None and self.text:
            tokenizer = Tokenizer()
            self.estimated_num_tokens = tokenizer.count_tokens(self.text)

    @classmethod
    def from_dict(cls, doc: Dict):
        """Create a Document object from a dictionary.

        Example:

        .. code-block :: python

            doc = Document.from_dict({
                "id": "123",
                "text": "Hello world",
                "meta_data": {"title": "Greeting"}
            })
        """

        doc = doc.copy()
        assert "meta_data" in doc, "meta_data is required"
        assert "text" in doc, "text is required"
        if "estimated_num_tokens" not in doc:
            tokenizer = Tokenizer()
            doc["estimated_num_tokens"] = tokenizer.count_tokens(doc["text"])
        if "id" not in doc or not doc["id"]:
            doc["id"] = uuid.uuid4()

        return super().from_dict(doc)

    def __repr__(self):
        """Custom repr method to truncate the text to 100 characters and vector to 10 floats."""
        max_chars_to_show = 100
        truncated_text = (
            self.text[:max_chars_to_show] + "..."
            if len(self.text) > max_chars_to_show
            else self.text
        )
        truncated_vector = (
            f"len: {len(self.vector)}" if len(self.vector) else self.vector
        )
        return (
            f"Document(id={self.id}, text={truncated_text!r}, meta_data={self.meta_data}, "
            f"vector={truncated_vector!r}, parent_doc_id={self.parent_doc_id}, order={self.order}, "
            f"score={self.score})"
        )


#######################################################################################
# Data modeling for dialog system
######################################################################################
@dataclass
class UserQuery:
    query_str: str
    metadata: Optional[Dict[str, Any]] = None


@dataclass
class AssistantResponse:
    response_str: str
    metadata: Optional[Dict[str, Any]] = None


# There could  more other roles in a multi-party conversation. We might consider in the future.
@dataclass
class DialogTurn(DataClass):
    __doc__ = r"""A turn consists of a user query and the assistant response.

    The dataformat is designed to fit into a relational database, where each turn is a row.
    Use `session_id` to group the turns into a dialog session with the `order` field and
    `user_query_timestamp` and `assistant_response_timestamp` to order the turns.

    Args:

        id (str): The unique id of the turn.
        user_id (str, optional): The unique id of the user.
        session_id (str, optional): The unique id of the dialog session.
        order (int, optional): The order of the turn in the dialog session, starts from 0.
        user_query (UserQuery, optional): The user query in the turn.
        assistant_response (AssistantResponse, optional): The assistant response in the turn.
        user_query_timestamp (datetime, optional): The timestamp of the user query.
        assistant_response_timestamp (datetime, optional): The timestamp of the assistant response.
        metadata (Dict[str, Any], optional): Additional metadata.

    Examples:

        - User: Hi, how are you?
        - Assistant: I'm fine, thank you!
        DialogTurn(id=uuid4(), user_query=UserQuery("Hi, how are you?"), assistant_response=AssistantResponse("I'm fine, thank you!"))
    """

    id: str = field(
        default_factory=lambda: str(uuid.uuid4()),
        metadata={"desc": "The unique id of the turn"},
    )
    user_id: Optional[str] = field(
        default=None, metadata={"desc": "The unique id of the user"}
    )
    conversation_id: Optional[str] = field(
        default=None,
        metadata={"desc": "The unique id of the conversation it belongs to"},
    )
    order: Optional[int] = field(
        default=None,
        metadata={"desc": "The order of the turn in the Dialog Session, starts from 0"},
    )

    user_query: Optional[UserQuery] = field(
        default=None, metadata={"desc": "The user query in the turn"}
    )
    assistant_response: Optional[AssistantResponse] = field(
        default=None, metadata={"desc": "The assistant response in the turn"}
    )
    user_query_timestamp: Optional[datetime] = field(
        default_factory=datetime.now,
        metadata={"desc": "The timestamp of the user query"},
    )
    assistant_response_timestamp: Optional[datetime] = field(
        default_factory=datetime.now,
        metadata={"desc": "The timestamp of the assistant response"},
    )
    metadata: Optional[Dict[str, Any]] = field(
        default=None, metadata={"desc": "Additional metadata"}
    )
    vector: Optional[List[float]] = field(
        default=None,
        metadata={"desc": "The vector representation of the dialog turn"},
    )

    def set_user_query(
        self, user_query: UserQuery, user_query_timestamp: Optional[datetime] = None
    ):
        self.user_query = user_query
        if not user_query_timestamp:
            user_query_timestamp = datetime.now()
        self.user_query_timestamp = user_query_timestamp

    def set_assistant_response(
        self,
        assistant_response: AssistantResponse,
        assistant_response_timestamp: Optional[datetime] = None,
    ):
        self.assistant_response = assistant_response
        if not assistant_response_timestamp:
            assistant_response_timestamp = datetime.now()
        self.assistant_response_timestamp = assistant_response_timestamp


# TODO: This part and the Memory class is still WIP, and will need more work in the future.
@dataclass
class Conversation:
    __doc__ = r"""A conversation manages the dialog turns in a whole conversation as a session.

    This class is mainly used in-memory for the dialog system/app to manage active conversations.
    You won't need this class for past conversations which have already been persisted in a database as a form of
    record or history.
    """

    id: str = field(
        default_factory=lambda: str(uuid.uuid4()),
        metadata={"desc": "The id of the conversation"},
    )  # the id of the conversation
    name: Optional[str] = field(
        default=None, metadata={"desc": "The name of the conversation"}
    )
    user_id: Optional[str] = field(
        default=None, metadata={"desc": "The id of the user"}
    )
    dialog_turns: OrderedDict[int, DialogTurn] = field(
        default_factory=OrderedDict, metadata={"desc": "The dialog turns"}
    )
    # int is the order of the turn, starts from 0
    metadata: Optional[Dict[str, Any]] = field(
        default=None, metadata={"desc": "Additional metadata"}
    )

    created_at: Optional[datetime] = field(
        default_factory=datetime.now,
        metadata={"desc": "The timestamp of the conversation creation"},
    )

    # InitVar type annotation is used for parameters that are used in __post_init__
    # but not meant to be fields in the dataclass.
    dialog_turns_input: InitVar[
        Optional[Union[OrderedDict[int, DialogTurn], List[DialogTurn]]]
    ] = None

    def __post_init__(
        self,
        dialog_turns_input: Optional[
            Union[OrderedDict[int, DialogTurn], List[DialogTurn]]
        ] = None,
    ):
        if dialog_turns_input:
            if isinstance(dialog_turns_input, list):
                # Assume the list is of DialogTurn objects and needs to be added to an OrderedDict
                for order, dialog_turn in enumerate(dialog_turns_input):
                    self.append_dialog_turn(dialog_turn)
            elif isinstance(dialog_turns_input, OrderedDict):
                self.dialog_turns = dialog_turns_input
            else:
                raise ValueError(
                    "dialog_turns should be a list of DialogTurn or an OrderedDict"
                )

    def get_next_order(self):
        return len(self.dialog_turns)

    def append_dialog_turn(self, dialog_turn: DialogTurn):
        next_order = self.get_next_order()
        if dialog_turn.order is None:
            dialog_turn.order = next_order
        else:
            assert dialog_turn.order == next_order, f"order should be {next_order}"
        self.dialog_turns[next_order] = dialog_turn

    def get_dialog_turns(self) -> OrderedDict[int, DialogTurn]:
        return self.dialog_turns

    def get_chat_history_str(self) -> str:
        chat_history_str = ""
        for order, dialog_turn in self.dialog_turns.items():
            chat_history_str += f"User: {dialog_turn.user_query.query_str}\n"
            chat_history_str += (
                f"Assistant: {dialog_turn.assistant_response.response_str}\n"
            )
        return chat_history_str

    def delete_dialog_turn(self, order: int):
        self.dialog_turns.pop(order)

    def update_dialog_turn(self, order: int, dialog_turn: DialogTurn):
        self.dialog_turns[order] = dialog_turn



"""A base class that provides an easy way for data to interact with LLMs."""

from typing import List, Dict, Any, Optional, Union, Callable, Type
import collections
from collections import OrderedDict


import enum
from copy import deepcopy
from dataclasses import (
    field,
    fields,
    make_dataclass,
    is_dataclass,
)

import json
import yaml
import logging

from adalflow.core.functional import (
    # dataclass_obj_to_dict,
    custom_asdict,
    dataclass_obj_from_dict,
    get_dataclass_schema,
    convert_schema_to_signature,
    represent_ordereddict,
)

__all__ = [
    "DataClass",
    "DataClassFormatType",
    "required_field",
    "ExcludeType",
    "IncludeType",
    "check_adal_dataclass",
    "DynamicDataClassFactory",
]
logger = logging.getLogger(__name__)


class DataClassFormatType(enum.Enum):
    r"""The format type for the DataClass schema."""

    # for class
    SCHEMA = "schema"
    SIGNATURE_YAML = "signature_yaml"
    SIGNATURE_JSON = "signature_json"
    # for instance
    EXAMPLE_YAML = "example_yaml"
    EXAMPLE_JSON = "example_json"


# Register the custom representer
yaml.add_representer(collections.OrderedDict, represent_ordereddict)


def required_field() -> Callable[[], Any]:
    """
    A factory function to create a required field in a dataclass.
    The returned callable raises a TypeError when invoked, indicating a required field was not provided.

    Args:
        name (Optional[str], optional): The name of the required field. Defaults to None

    Returns:
        Callable[[], Any]: A callable that raises TypeError when called, indicating a missing required field.

    Example:

    .. code-block:: python

        from dataclasses import dataclass
        from adalflow.core.base_data_class import required_field, DataClass

        @dataclass
        class Person(DataClass):
            name: str = field(default=None)
            age: int = field(default_factory=required_field())# allow required field after optional field
    """

    def required_field_error():
        """This function is returned by required_field and raises an error indicating the field is required."""
        raise TypeError("This field is required and was not provided.")

    required_field_error.__name__ = (
        "required_field"  # Set the function's name explicitly
    )
    return required_field_error


# Dict is for the nested dataclasses, e.g. {"Person": ["name", "age"], "Address": ["city"]}
ExcludeType = Optional[Union[List[str], Dict[str, List[str]]]]
IncludeType = Optional[Union[List[str], Dict[str, List[str]]]]


class DataClass:
    __doc__ = r"""The base data class for all data types that interact with LLMs.

    Please only exclude optional fields in the exclude dictionary.

    Designed to streamline the handling, serialization, and description of data within our applications, especially to LLM prompt.
    We explicitly handle this instead of relying on 3rd party libraries such as pydantic or marshmallow to have better
    transparency and to keep the order of the fields when get serialized.

    How to create your own dataclass?

    1. Subclass DataClass and define the fields with the `field` decorator.
    2. Use the `medata` argument and a `desc` key to describe the field.
    3. Keep the order of the fields as how you want them to be serialized and described to LLMs.
    4. field with default value is considered optional. Field without default value and field with default_factory=required_field is considered required.

    How to use it?

    Describing:

    We defined :class:`DataClassFormatType<core.types.DataClassFormatType>` to categorize DataClass description formats
    as input or output in LLM prompt.


    (1) For describing the class (data structure):

    `Signature` is more token effcient than schema, and schema as it is always a json string, when you want LLMs to output yaml, it can be misleading if you describe the data structure in json.

    - DataClassFormatType.SCHEMA: a more standard way to describe the data structure in Json string, :meth:`to_schema` as string and :meth:`to_schema` as dict.
    - DataClassFormatType.SIGNATURE_JSON: emitating a json object with field name as key and description as value, :meth:`to_json_signature` as string.
    - DataClassFormatType.SIGNATURE_YAML: emitating a yaml object with field name as key and description as value, :meth:`to_yaml_signature` as string.

    (2) For describing the class instance: this is helpful to do few-shot examples in LLM prompt.
    - DataClassFormatType.EXAMPLE_JSON: the json representation of the instance, :meth:`to_json` as string.
    - DataClassFormatType.EXAMPLE_YAML: the yaml representation of the instance, :meth:`to_yaml` as string.

    Overall, we have a unified class method :meth:`format_str` to generate formatted output based on the type of operation and class/instance context.

    note::
        1. Avoid using Optional[Type] for the type of fields, as dataclass already distingushes between optional and required fields using default value.
        2. If you need to customize, you can subclass and overwrite any method to fit your needs.

    Loading data:

    - :meth:`from_dict` is used to create a dataclass instance from a dictionary.


    Refer :ref:`DataClass<core-base_data_class_note>` for more detailed instructions.

    Examples:

    .. code-block:: python

        # Define a dataclass
        from adalflow.core import DataClass
        from dataclasses import dataclass, field

        @dataclass
        class MyOutputs(DataClass):
            age: int = field(metadata={"desc": "The age of the person", "prefix": "Age:"})
            name: str = field(metadata={"desc": "The name of the person", "prefix": "Name:"})

        # Create json signature
        print(MyOutputs.to_json_signature())
        # Output:
        # {
        #     "age": "The age of the person",
        #     "name": "The name of the person"
        # }
        # Create yaml signature
        print(MyOutputs.to_yaml_signature())
        # Output:
        # age: The age of the person
        # name: The name of the person

        # Create a dataclass instance
        my_instance = MyOutputs(age=25, name="John Doe")
        # Create json example
        print(my_instance.to_json_example())
        # Output:
        # {
        #     "age": 25,
        #     "name": "John Doe"
        # }
        # Create yaml signature
        print(my_instance.to_yaml_example())
        # Output:
        # age: 25
        # name: John Doe

    """
    __input_fields__: List[str] = []
    __output_fields__: List[str] = []

    def __post_init__(self):

        for f in fields(self):
            if "desc" not in f.metadata and "description" not in f.metadata:

                logger.debug(
                    f"Class {  self.__class__.__name__} Field {f.name} is missing 'desc' in metadata"
                )

    @classmethod
    def get_task_desc(cls) -> str:
        """Get the task description for the dataclass.

        Returns:
            str: The task description for the dataclass.
        """
        return cls.__doc__

    @classmethod
    def set_task_desc(cls, task_desc: str) -> None:
        """Set the task description for the dataclass.

        Args:
            task_desc (str): The task description to set.
        """
        cls.__doc__ = task_desc

    @classmethod
    def get_input_fields(cls):
        """Return a list of all input fields."""
        return cls.__input_fields__

    @classmethod
    def set_input_fields(cls, input_fields: List[str]):
        """Set the input fields for the dataclass.
          When creating schema or instance, it will follow the input field and output field order

        Args:
            input_fields (List[str]): The input fields to set.
        """
        cls.__input_fields__ = input_fields

    @classmethod
    def get_output_fields(cls):
        """Return a list of all output fields."""
        return cls.__output_fields__

    @classmethod
    def set_output_fields(cls, output_fields: List[str]):
        """Set the output fields for the dataclass.
          When creating schema or instance, it will follow the input field and output field order

        Args:
            output_fields (List[str]): The output fields to set.
        """
        cls.__output_fields__ = output_fields

    def to_dict(
        self,
        *,
        exclude: ExcludeType = None,
        include: IncludeType = None,
    ) -> Dict[str, Any]:
        """Convert a dataclass object to a dictionary.

        Supports nested dataclasses, lists, and dictionaries.
        Allow exclude keys for each dataclass object.

        Use cases:
        - Decide what information will be included to be serialized to JSON or YAML that can be used in LLM prompt.
        - Exclude sensitive information from the serialized output.
        - Serialize the dataclass instance to a dictionary for saving states.

        Args:
            exclude (Optional[Dict[str, List[str]]], optional): A dictionary of fields to exclude for each dataclass object. Defaults to None.


        Example:

        .. code-block:: python

            from dataclasses import dataclass
            from typing import List

            @dataclass
            class TrecData:
                question: str
                label: int

            @dataclass
            class TrecDataList(DataClass):

                data: List[TrecData]
                name: str

            trec_data = TrecData(question="What is the capital of France?", label=0)
            trec_data_list = TrecDataList(data=[trec_data], name="trec_data_list")

            trec_data_list.to_dict(exclude={"TrecData": ["label"], "TrecDataList": ["name"]})

            # Output:
            # {'data': [{'question': 'What is the capital of France?'}]}
        """
        if not is_dataclass(self):
            raise ValueError(
                f"to_dict() is not called on a dataclass instance: {self.__class__}. You might forget to use @dataclass decorator."
            )
        # convert all fields to its data if its parameter
        fields = self.__dataclass_fields__
        from adalflow.optim.parameter import Parameter

        for f in fields.values():
            # print(f"field: {f}")
            field_value = getattr(self, f.name)
            # if its a parameter, convert to its data
            if isinstance(field_value, Parameter):
                setattr(self, f.name, field_value.data)
        # print(f"adapted self: {self}")

        # ensure only either include or exclude is used not both
        if include and exclude:
            raise ValueError("Either include or exclude can be used, not both.")

        # convert include to excluded

        excluded: Optional[Dict[str, List[str]]] = None
        if include:  # only support unnested fields
            # fild all fields of the class
            fields = self.__dataclass_fields__
            # generate the excluded dict
            excluded = {
                self.__class__.__name__: [
                    f.name for f in fields.values() if f.name not in include
                ]
            }
        elif exclude:
            if exclude and isinstance(exclude, List):
                excluded = {self.__class__.__name__: exclude}
            elif exclude and isinstance(exclude, Dict):
                excluded = deepcopy(exclude)
            else:
                excluded = None
        # return custom_asdict(self, exclude=excluded)
        # Convert the dataclass to a dictionary

        raw_dict = custom_asdict(self, exclude=excluded)

        # Reorder the dictionary based on input_field and output_field
        input_fields = self.get_input_fields()
        output_fields = self.get_output_fields()

        ordered_dict = OrderedDict()

        # First, add input fields in order
        for field_name in input_fields:
            if field_name in raw_dict:
                ordered_dict[field_name] = raw_dict[field_name]

        # Then, add output fields in order
        for field_name in output_fields:
            if field_name in raw_dict:
                ordered_dict[field_name] = raw_dict[field_name]

        # Finally, add any remaining fields (if there are any)
        for field_name, value in raw_dict.items():
            if field_name not in ordered_dict:
                ordered_dict[field_name] = value

        return dict(ordered_dict)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "DataClass":
        """Create a dataclass instance from a dictionary.

        Supports nested dataclasses, lists, and dictionaries.

        Example from the :meth:`to_dict` method.

        ..code-block:: python

            data_dict = trec_data_list.to_dict()
            restored_data = TreDataList.from_dict(data_dict)

            assert str(restored_data.__dict__) == str(trec_data_list.__dict__)

        .. note::
        If any required field is missing, it will raise an error.
        Do not use the dict that has excluded required fields.

        Use cases:
        - Convert the json/yaml output from LLM prediction to a dataclass instance.
        - Restore the dataclass instance from the serialized output used for states saving.
        """
        try:
            dclass = dataclass_obj_from_dict(cls, data)
            logger.debug(f"Dataclass instance created from dict: {dclass}")
            return dclass
        except TypeError as e:
            raise ValueError(f"Failed to load data: {e}")

    @classmethod
    def from_json(cls, json_str: str) -> "DataClass":
        """Create a dataclass instance from a JSON string.

        Args:
            json_str (str): The JSON string to convert to a dataclass instance.

        Example:

        .. code-block:: python

            json_str = '{"question": "What is the capital of France?", "label": 0}'
            trec_data = TrecData.from_json(json_str)
        """
        try:
            data = json.loads(json_str)
            return cls.from_dict(data)
        except json.JSONDecodeError as e:
            raise ValueError(f"Failed to load JSON string: {e}")

    def to_json_obj(
        self,
        exclude: ExcludeType = None,
        include: IncludeType = None,
    ) -> Any:
        r"""Convert the dataclass instance to a JSON object.

        :meth:`to_dict` along with the use of sort_keys=False to ensure the order of the fields is maintained.
        This can be important to llm prompt.

        Args:

        exclude (Optional[Dict[str, List[str]]], optional): A dictionary of fields to exclude for each dataclass object. Defaults to None.
        """
        return json.loads(self.to_json(exclude=exclude, include=include))

    def to_json(
        self,
        exclude: ExcludeType = None,
        include: IncludeType = None,
    ) -> str:
        r"""Convert the dataclass instance to a JSON string.

        :meth:`to_dict` along with the use of sort_keys=False to ensure the order of the fields is maintained.
        This can be important to llm prompt.

        Args:

        exclude (Optional[Dict[str, List[str]]], optional): A dictionary of fields to exclude for each dataclass object. Defaults to None.
        """
        return json.dumps(
            self.to_dict(exclude=exclude, include=include), indent=4, sort_keys=False
        )

    @classmethod
    def from_yaml(cls, yaml_str: str) -> "DataClass":
        """Create a dataclass instance from a YAML string.

        Args:
            yaml_str (str): The YAML string to convert to a dataclass instance.

        Example:

        .. code-block:: python

            yaml_str = 'question: What is the capital of France?\nlabel: 0'
            trec_data = TrecData.from_yaml(yaml_str)
        """
        try:
            data = yaml.safe_load(yaml_str)
            return cls.from_dict(data)
        except yaml.YAMLError as e:
            raise ValueError(f"Failed to load YAML string: {e}")

    def to_yaml_obj(
        self,
        exclude: ExcludeType = None,
        include: IncludeType = None,
    ) -> Any:
        r"""Convert the dataclass instance to a YAML object.

        :meth:`to_dict` along with the use of sort_keys=False to ensure the order of the fields is maintained.

        Args:

        exclude (Optional[Dict[str, List[str]]], optional): A dictionary of fields to exclude for each dataclass object. Defaults to None.
        """
        return yaml.safe_load(self.to_yaml(exclude=exclude, include=include))

    def to_yaml(
        self,
        exclude: ExcludeType = None,
        include: IncludeType = None,
    ) -> str:
        r"""Convert the dataclass instance to a YAML string.

        :meth:`to_dict` along with the use of sort_keys=False to ensure the order of the fields is maintained.

        Args:

        exclude (Optional[Dict[str, List[str]]], optional): A dictionary of fields to exclude for each dataclass object. Defaults to None.
        """
        return yaml.dump(
            self.to_dict(exclude=exclude, include=include),
            default_flow_style=False,
            sort_keys=False,
        ).strip()

    def dict_to_yaml(self, data: Dict[str, Any]) -> str:
        """Convert a dictionary to a YAML string.

        Args:
            data (Dict[str, Any]): The dictionary to convert to a YAML string.

        Returns:
            str: The YAML string representation of the dictionary.
        """
        return yaml.dump(data, default_flow_style=False, sort_keys=False).strip()

    @classmethod
    def to_schema(
        cls,
        exclude: ExcludeType = None,
        include: IncludeType = None,
    ) -> Dict[str, Dict[str, Any]]:
        """Generate a Json schema which is more detailed than the signature."""
        # convert exclude to dict if it is a list
        if include and exclude:
            raise ValueError("Either include or exclude can be used, not both.")
        excluded: Optional[Dict[str, List[str]]] = None
        if include:  # only support unnested fields
            # fild all fields of the class
            fields = cls.__dataclass_fields__
            # generate the excluded dict
            excluded = {
                cls.__name__: [f.name for f in fields.values() if f.name not in include]
            }
        elif exclude:
            if exclude and isinstance(exclude, List):
                excluded = {cls.__name__: exclude}
            elif exclude and isinstance(exclude, Dict):
                excluded = deepcopy(exclude)
            else:
                excluded = None

        raw_dict = get_dataclass_schema(
            cls, excluded, getattr(cls, "__type_var_map__", None)
        )
        # Reorder the dictionary based on input_field and output_field

        properties = raw_dict.get("properties", {})
        # reorder the properties fields
        input_fields = cls.get_input_fields()
        output_fields = cls.get_output_fields()

        ordered_dict = OrderedDict()

        # First, add input fields in order
        for field_name in input_fields:
            if field_name in properties:
                ordered_dict[field_name] = properties[field_name]

        # Then, add output fields in order
        for field_name in output_fields:
            if field_name in properties:
                ordered_dict[field_name] = properties[field_name]

        # Finally, add any remaining fields (if there are any)
        for field_name, value in properties.items():
            if field_name not in ordered_dict:
                ordered_dict[field_name] = value

        # Update the properties field
        raw_dict["properties"] = ordered_dict

        return raw_dict

    @classmethod
    def to_schema_str(
        cls,
        exclude: ExcludeType = None,
        include: IncludeType = None,
    ) -> str:
        """Generate a Json schema which is more detailed than the signature."""
        schema = cls.to_schema(exclude=exclude, include=include)
        return json.dumps(schema, indent=4).strip()

    @classmethod
    def to_yaml_signature(
        cls,
        exclude: ExcludeType = None,
        include: IncludeType = None,
    ) -> str:
        r"""Generate a YAML signature for the class from desc in metadata.

        Used mostly as LLM prompt to describe the output data format.
        """
        # NOTE: we manually format the yaml string as the yaml.dump will always sort the keys
        # Which can impact the final model output
        schema = cls.to_schema(exclude=exclude, include=include)
        signature_dict = convert_schema_to_signature(schema)
        yaml_content = []
        for key, value in signature_dict.items():
            yaml_content.append(f"{key}: {value}")

        yaml_output = "\n".join(yaml_content)
        return yaml_output

    @classmethod
    def to_json_signature(
        cls,
        exclude: ExcludeType = None,
        include: IncludeType = None,
    ) -> str:
        """Generate a JSON `signature`(json string) for the class from desc in metadata.

        Used mostly as LLM prompt to describe the output data format.

        Example:

        >>> @dataclass
        >>> class MyOutputs(DataClass):
        >>>    age: int = field(metadata={"desc": "The age of the person", "prefix": "Age:"})
        >>>    name: str = field(metadata={"desc": "The name of the person", "prefix": "Name:"})

        >>> print(MyOutputs.to_json_signature())
        >>> # Output is a JSON string:
        >>> # '{
        >>> #    "age": "The age of the person (int) (required)",
        >>> #    "name": "The name of the person (str) (required)"
        >>> #}'
        """
        schema = cls.to_schema(exclude=exclude, include=include)
        signature_dict = convert_schema_to_signature(schema)

        return json.dumps(signature_dict, indent=4)

    @classmethod
    def to_dict_class(
        cls,
        exclude: ExcludeType = None,
        include: IncludeType = None,
    ) -> Dict[str, Any]:
        """More of an internal used class method for serialization.

        Converts the dataclass to a dictionary, optionally excluding specified fields.
        Use this to save states of the class in serialization, not advised to use in LLM prompt.
        """
        return cls.to_schema(exclude=exclude, include=include)

    @classmethod
    def format_class_str(
        cls,
        format_type: DataClassFormatType,
        exclude: ExcludeType = None,
        include: IncludeType = None,
    ) -> str:
        """Generate formatted output based on the type of operation and class/instance context.

        Args:
            format_type (DataClassFormatType): Specifies the format and type (schema, signature, example).

        Returns:
            str: A string representing the formatted output.

        Examples:

        .. code-block:: python

            # Define a dataclass
            from adalflow.core import DataClass

        """

        if format_type == DataClassFormatType.SIGNATURE_JSON:
            return cls.to_json_signature(exclude=exclude, include=include)
        elif format_type == DataClassFormatType.SIGNATURE_YAML:
            return cls.to_yaml_signature(exclude=exclude, include=include)

        elif format_type == DataClassFormatType.SCHEMA:
            return cls.to_schema_str(exclude=exclude, include=include)
        else:
            raise ValueError(f"Unsupported format type: {format_type}")

    def format_example_str(
        self,
        format_type: DataClassFormatType,
        exclude: ExcludeType = None,
        include: IncludeType = None,
    ) -> str:
        """Generate formatted output based on the type of operation and class/instance context.

        Args:
            format_type (DataClassFormatType): Specifies the format and type (schema, signature, example).

        Returns:
            str: A string representing the formatted output.

        """

        # Check the type of format required and whether it's called on an instance or class
        if format_type == DataClassFormatType.EXAMPLE_JSON:
            return self.to_json(exclude=exclude, include=include)
        elif format_type == DataClassFormatType.EXAMPLE_YAML:
            return self.to_yaml(exclude=exclude, include=include)
        else:
            raise ValueError(f"Unsupported format type: {format_type}")


def check_adal_dataclass(data_class: Type) -> None:
    """Check if the provided class is a valid dataclass for the AdalFlow framework.

    Args:
        data_class (Type): The class to check.
    """

    if not is_dataclass(data_class):
        raise TypeError(f"Provided class is not a dataclass: {data_class}")

    if not issubclass(data_class, DataClass):
        raise TypeError(f"Provided class is not a subclass of DataClass: {data_class}")


class DynamicDataClassFactory:
    @staticmethod
    def from_dict(
        data: Dict[str, Any],
        base_class: Type = DataClass,
        class_name: str = "DynamicDataClass",
    ) -> DataClass:
        """
        Create an instance of a dataclass from a dictionary. The dictionary should have the following structure:
        {
            "field_name": field_value,
            ...
        }

        Args:
            data (dict): The dictionary with field names and values.
            base_class (type): The base class to inherit from (default: BaseDataClass).
            class_name (str): The name of the generated dataclass (default: DynamicDataClass).

        Returns:
            BaseDataClass: An instance of the generated dataclass.
        """
        # Create field specifications for the dataclass
        # fields_spec = [
        #     (key, type(value), field(default=value)) for key, value in data.items()
        # ]
        fields_spec = []
        for key, value in data.items():
            field_type = type(value)
            if isinstance(value, (list, dict, set)):
                fields_spec.append(
                    (key, field_type, field(default_factory=lambda v=value: v))
                )
            else:
                fields_spec.append((key, field_type, field(default=value)))

        # Create the dataclass
        dynamic_class = make_dataclass(class_name, fields_spec, bases=(base_class,))

        # Create an instance of the dataclass with the provided values
        instance = dynamic_class.from_dict(data)

        return instance



"""Generator is a user-facing orchestration component with a simple and unified interface for LLM prediction.

It is a pipeline that consists of three subcomponents."""

import json
import re
import os
from pathlib import Path

from typing import Any, Dict, Optional, Union, Callable, Tuple, List
import logging
from dataclasses import dataclass, field


from adalflow.core.types import (
    ModelType,
    GeneratorOutput,
    GeneratorOutputType,
)
from adalflow.core.component import Component, DataComponent
from adalflow.optim.grad_component import GradComponent
from adalflow.core.base_data_class import DataClass


from adalflow.optim.parameter import (
    Parameter,
    OutputParameter,
)
from adalflow.optim.gradient import GradientContext, Gradient
from adalflow.optim.types import ParameterType

from adalflow.core.prompt_builder import Prompt
from adalflow.core.functional import compose_model_kwargs
from adalflow.core.model_client import ModelClient
from adalflow.core.default_prompt_template import DEFAULT_ADALFLOW_SYSTEM_PROMPT
from adalflow.optim.function import BackwardContext
from adalflow.utils.cache import CachedEngine
from adalflow.tracing.callback_manager import CallbackManager
from adalflow.utils.global_config import get_adalflow_default_root_path
from adalflow.core.string_parser import JsonParser


from adalflow.optim.text_grad.backend_engine_prompt import (
    FEEDBACK_ENGINE_TEMPLATE,
    LLM_CONVERSATION_TEMPLATE,
    ALL_PRED_INFO,
    OUTPUT_INSTRUCTION,
    VARIABLE_AND_PEERS_INFO,
    CONVERSATION_START_INSTRUCTION_CHAIN,
    OBJECTIVE_INSTRUCTION_BASE,
    OBJECTIVE_INSTRUCTION_CHAIN,
)
from adalflow.utils.logger import printc

__all__ = ["Generator", "BackwardEngine", "create_teacher_generator"]


log = logging.getLogger(__name__)

DEBUG_MODE = os.environ.get("DEBUG_MODE", False)

PromptArgType = Dict[str, Union[str, Parameter]]


@dataclass
class BackwardPassSetup(DataClass):
    all_pred_at_once: bool = field(
        default=False, metadata={"desc": "Backward all predecessors at once."}
    )
    threshold_score_to_compute_grad_for_errors: float = field(
        default=0.9,
        metadata={"desc": "Threshold score to compute gradient for errors."},
    )
    compute_grad_for_errors_only: bool = field(
        default=True, metadata={"desc": "Compute gradient for errors only."}
    )


class Generator(GradComponent, CachedEngine, CallbackManager):
    __doc__ = """An user-facing orchestration component for LLM prediction.

    It is also a GradComponent that can be used for backpropagation through the LLM model.

    By orchestrating the following three components along with their required arguments,
    it enables any LLM prediction with required task output format.
    - Prompt
    - Model client
    - Output processors

    Args:
        model_client (ModelClient): The model client to use for the generator.
        model_kwargs (Dict[str, Any], optional): The model kwargs to pass to the model client. Defaults to {}. Please refer to :ref:`ModelClient<components-model_client>` for the details on how to set the model_kwargs for your specific model if it is from our library.
        template (Optional[str], optional): The template for the prompt.  Defaults to :ref:`DEFAULT_ADALFLOW_SYSTEM_PROMPT<core-default_prompt_template>`.
        prompt_kwargs (Optional[Dict], optional): The preset prompt kwargs to fill in the variables in the prompt. Defaults to None.
        output_processors (Optional[Component], optional):  The output processors after model call. It can be a single component or a chained component via ``Sequential``. Defaults to None.
        trainable_params (Optional[List[str]], optional): The list of trainable parameters. Defaults to [].

    Note:
        The output_processors will be applied to the string output of the model completion. And the result will be stored in the data field of the output.
        And we encourage you to only use it to parse the response to data format you will use later.
    """

    model_type: ModelType = ModelType.LLM
    model_client: ModelClient  # for better type checking

    _use_cache: bool = False
    _kwargs: Dict[str, Any] = (
        {}
    )  # to create teacher generator from student TODO: might reaccess this

    backward_pass_setup: BackwardPassSetup = (
        BackwardPassSetup()
    )  # default setup for the backward pass

    def __init__(
        self,
        *,
        # args for the model
        model_client: ModelClient,  # will be intialized in the main script
        model_kwargs: PromptArgType = {},
        # args for the prompt
        template: Optional[str] = None,
        prompt_kwargs: Optional[Dict] = {},
        # args for the output processing
        output_processors: Optional[DataComponent] = None,
        name: Optional[str] = None,
        # args for the cache
        cache_path: Optional[str] = None,
        use_cache: bool = False,
    ) -> None:
        r"""The default prompt is set to the DEFAULT_ADALFLOW_SYSTEM_PROMPT. It has the following variables:
        - task_desc_str
        - tools_str
        - example_str
        - chat_history_str
        - context_str
        - steps_str
        You can preset the prompt kwargs to fill in the variables in the prompt using prompt_kwargs.
        But you can replace the prompt and set any variables you want and use the prompt_kwargs to fill in the variables.
        """

        if not isinstance(model_client, ModelClient):
            raise TypeError(
                f"{type(self).__name__} requires a ModelClient instance for model_client, please pass it as OpenAIClient() or GroqAPIClient() for example.\
                    Got {model_client} instead."
            )

        template = template or DEFAULT_ADALFLOW_SYSTEM_PROMPT

        # create the cache path and initialize the cache engine

        self.set_cache_path(
            cache_path, model_client, model_kwargs.get("model", "default")
        )

        CachedEngine.__init__(self, cache_path=self.cache_path)

        Component.__init__(self)
        GradComponent.__init__(self, desc="Generate a response using LLM model.")
        CallbackManager.__init__(self)

        self.name = name or self.__class__.__name__

        self._init_prompt(template, prompt_kwargs)

        self.model_kwargs = model_kwargs.copy()
        # init the model client
        self.model_client = model_client

        self.output_processors = output_processors

        if output_processors and (not isinstance(output_processors, DataComponent)):
            raise ValueError(
                f"output_processors should be a DataComponent instance, got {type(output_processors)}"
            )

        self.set_parameters(prompt_kwargs)

        # end of trainable parameters
        self.backward_engine: "BackwardEngine" = None
        log.info(f"Generator {self.name} initialized.")
        #  to support better testing on the parts beside of the model call
        self.mock_output: bool = False
        self.mock_output_data: str = "mock data"

        self._use_cache = use_cache

        self._kwargs = {
            "model_client": model_client,
            "model_kwargs": model_kwargs,
            "template": template,
            "prompt_kwargs": prompt_kwargs,
            "output_processors": output_processors,
            "name": name,
            "cache_path": cache_path,
            "use_cache": use_cache,
        }
        self._teacher: Optional["Generator"] = None
        self._trace_api_kwargs: Dict[str, Any] = (
            {}
        )  # used by dynamic computation graph and backpropagation

    def update_default_backward_pass_setup(self, setup: BackwardPassSetup):
        self.backward_pass_setup = setup

    def set_cache_path(self, cache_path: str, model_client: object, model: str):
        """Set the cache path for the generator."""

        # Construct a valid model string using the client class name and model
        self.model_str = f"{model_client.__class__.__name__}_{model}"

        # Remove any characters that are not allowed in file names (cross-platform)
        # On Windows, characters like `:<>?/\|*` are prohibited.
        self.model_str = re.sub(r"[^a-zA-Z0-9_\-]", "_", self.model_str)

        _cache_path = (
            get_adalflow_default_root_path() if cache_path is None else cache_path
        )

        # Use pathlib to handle paths more safely across OS
        self.cache_path = Path(_cache_path) / f"cache_{self.model_str}.db"

        log.debug(f"Cache path set to: {self.cache_path}")

    def get_cache_path(self) -> str:
        r"""Get the cache path for the generator."""
        return self.cache_path

    @staticmethod
    def _get_default_mapping(
        output: "GeneratorOutput" = None,
    ) -> Tuple[Dict[str, Callable], List[str]]:

        if (
            output.data
            and isinstance(output.data, DataClass)
            and len(output.data.get_output_fields()) > 0
        ):
            output_fields = output.data.get_output_fields()

            output_mapping = {
                f: lambda x, f=f: getattr(x.data, f) for f in output_fields
            }
        elif output.raw_response:
            output_fields = ["raw_response"]
            output_mapping = {f: lambda x, f=f: getattr(x, f) for f in output_fields}
            output_fields = ["Answer"]
            output_mapping["Example"] = output_mapping["raw_response"]
            del output_mapping["raw_response"]

        return output_mapping, output_fields

    def set_mock_output(
        self, mock_output: bool = True, mock_output_data: str = "mock data"
    ):
        self.mock_output = mock_output
        self.mock_output_data = mock_output_data

    def reset_mock_output(self):
        self.mock_output = False
        self.mock_output_data = "mock data"

    def set_parameters(self, prompt_kwargs: PromptArgType):
        r"""Set name for each paramter and set all context for each other.
        Make all parameters attributes to the generator for finding them easily
        for optimizers and other components.
        """
        for key, p in prompt_kwargs.items():
            if isinstance(p, Parameter):
                if not p.name or p.name == "":
                    p.name = key
                peers = [
                    p
                    for k, p in prompt_kwargs.items()
                    if isinstance(p, Parameter) and k != key
                    # and p.param_type == ParameterType.PROMPT
                ]
                p.set_peers(peers)
                setattr(self, key, p)

    def _init_prompt(self, template: str, prompt_kwargs: Dict):
        r"""Initialize the prompt with the template and prompt_kwargs."""
        self.template = template
        self.prompt_kwargs = prompt_kwargs
        # NOTE: Prompt can handle parameters
        self.prompt = Prompt(template=template, prompt_kwargs=self.prompt_kwargs)

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "Generator":
        r"""Create a Generator instance from the config dictionary.

        Example:

        .. code-block:: python

            config = {
                        "model_client": {
                            "component_name": "OpenAIClient",
                            "component_config": {}
                        },
                        "model_kwargs": {"model": "gpt-3.5-turbo", "temperature": 0}
                    }
            generator = Generator.from_config(config)
        """
        # create init_kwargs from the config
        assert "model_client" in config, "model_client is required in the config"
        return super().from_config(config)

    def _compose_model_kwargs(self, **model_kwargs) -> Dict:
        r"""
        The model configuration exclude the input itself.
        Combine the default model, model_kwargs with the passed model_kwargs.
        Example:
        model_kwargs = {"temperature": 0.5, "model": "gpt-3.5-turbo"}
        self.model_kwargs = {"model": "gpt-3.5-turbo"}
        combine_kwargs(model_kwargs) => {"temperature": 0.5, "model": "gpt-3.5-turbo"}

        """
        combined_model_kwargs = self.model_kwargs.copy()

        if model_kwargs:
            combined_model_kwargs.update(model_kwargs)
        return combined_model_kwargs

    # TODO: use prompt_kwargs as users are already familiar with it
    def print_prompt(self, **kwargs) -> str:
        return self.prompt.print_prompt(**kwargs)

    def get_prompt(self, **kwargs) -> str:
        return self.prompt.call(**kwargs)

    def _extra_repr(self) -> str:
        s = f"model_kwargs={self.model_kwargs}, model_type={self.model_type}, prompt={self.prompt}"
        return s

    def _post_call(self, completion: Any) -> GeneratorOutput:
        r"""Get string completion and process it with the output_processors."""
        # parse chat completion will only fill the raw_response
        output: GeneratorOutput = self.model_client.parse_chat_completion(completion)
        # Now adding the data filed to the output
        data = output.raw_response
        if self.output_processors:
            if data:
                try:
                    data = self.output_processors(data)
                    output.data = data
                except Exception as e:
                    log.error(f"Error processing the output processors: {e}")
                    output.error = str(e)

        else:
            output.data = data

        return output

    def _pre_call(self, prompt_kwargs: Dict, model_kwargs: Dict) -> Dict[str, Any]:
        r"""Prepare the input, prompt_kwargs, model_kwargs for the model call."""
        # 1. render the prompt from the template
        prompt_str = self.prompt.call(**prompt_kwargs).strip()

        # 2. combine the model_kwargs with the default model_kwargs
        composed_model_kwargs = self._compose_model_kwargs(**model_kwargs)

        # 3. convert app's inputs to api inputs
        api_kwargs = self.model_client.convert_inputs_to_api_kwargs(
            input=prompt_str,
            model_kwargs=composed_model_kwargs,
            model_type=self.model_type,
        )
        # printc(f"api_kwargs: {api_kwargs}", color="red")
        return api_kwargs

    def _model_client_call(self, api_kwargs: Dict, use_cache: bool = False) -> Any:
        # call the model client
        try:
            # check the cache
            index_content = json.dumps(api_kwargs)  # + f"training: {self.training}"
            if use_cache:
                # print(f"check cache first: {no_cache}")

                cached_completion = self._check_cache(index_content)
                if cached_completion is not None:
                    return cached_completion

            completion = self.model_client.call(
                api_kwargs=api_kwargs, model_type=self.model_type
            )
            # prepare cache
            if use_cache:
                self._save_cache(index_content, completion)
            return completion
        except Exception as e:
            log.error(f"Error calling the model: {e}")
            raise e

    ##############################################################################################################
    ### Forward, backwards, teacher generator, create demo data instance,
    # are for training and backpropagation
    ##############################################################################################################

    def create_demo_data_instance(
        self,
        input_prompt_kwargs: Dict[str, Any],
        output: GeneratorOutput,
        id: Optional[str] = None,
    ):
        r"""Automatically create a demo data instance from the input and output of the generator.
        Used to trace the demos for the demo paramter in the prompt_kwargs.
        Part of the few-shot learning.
        """
        from adalflow.core.base_data_class import DynamicDataClassFactory

        # map the input fields
        demo_data = {"id": id, "score": None}  # add score to trace the prediction score
        demo_data_class_output_mapping, output_fields = self._get_default_mapping(
            output
        )

        for k, v in input_prompt_kwargs.items():
            if isinstance(v, Parameter):
                demo_data[k] = v.map_to_successor(self)
            else:
                demo_data[k] = v
        # map the output fields
        for key, value in demo_data_class_output_mapping.items():
            demo_data[key] = value(output)

        obj = DynamicDataClassFactory.from_dict(demo_data)
        obj.set_input_fields([k for k in input_prompt_kwargs.keys()])
        obj.set_output_fields(output_fields)
        if obj is None:
            raise ValueError(f"Error creating the demo data instance:{demo_data}")
        return obj

    def set_backward_engine(self, backward_engine: "BackwardEngine" = None):
        if backward_engine is None:
            backward_engine = BackwardEngine(
                model_client=self.model_client,
                model_kwargs=self.model_kwargs,
            )
            if self.mock_output:
                backward_engine.set_mock_output()
        self.backward_engine = backward_engine

    def set_teacher_generator(self, teacher: "Generator" = None):
        self._teacher = teacher
        print(f"Teacher generator set: {self._teacher}, teacher {teacher}")
        log.debug(f"Teacher generator set: {self._teacher}")

    # def set_data_map_func(self, map_func: Callable = None):
    #     def default_map_func(data: "GeneratorOutputType") -> str:
    #         return (
    #             data.data
    #             if data.data
    #             else self.failure_message_to_backward_engine(data)
    #         )

    #     self.data_map_func = map_func or default_map_func

    #     log.debug(f"Data map function set: {self.data_map_func}")

    # TODO: limit to only one demo parameter.
    @staticmethod
    def find_demo_parameter(prompt_kwargs: Dict) -> Optional[Parameter]:
        from adalflow.optim.parameter import Parameter, ParameterType

        for p in prompt_kwargs.values():
            if isinstance(p, Parameter) and p.param_type == ParameterType.DEMOS:
                return p
        return None

    def forward(
        self,
        prompt_kwargs: Optional[
            Dict[str, Union[str, Parameter]]
        ] = {},  # the input need to be passed to the prompt
        model_kwargs: Optional[Dict] = {},
        id: Optional[str] = None,
    ) -> "Parameter":
        r"""Customized forward pass on top of the GradComponent forward method."""
        # 1. convert prompt_kwargs to parameter if it is not
        for k, v in prompt_kwargs.items():
            if not isinstance(v, Parameter):
                prompt_kwargs[k] = Parameter(
                    data=v,
                    name=f"{self.name}_{k}",
                    requires_opt=False,
                    param_type=ParameterType.INPUT,
                    data_id=id,
                )

        # 2. call the model
        unwrapped_prompt_kwargs: Dict[str, Any] = {}
        for k, v in prompt_kwargs.items():
            if isinstance(v, Parameter):
                if v.param_type == ParameterType.INPUT:
                    v.data_id = id
                unwrapped_prompt_kwargs[k] = v.map_to_successor(self)
            else:
                unwrapped_prompt_kwargs[k] = v
        if DEBUG_MODE:
            print(
                f"unwrapped_prompt_kwargs: {unwrapped_prompt_kwargs}, model_kwargs: {model_kwargs}"
            )
            print(f"prompt template: {self.template}")

        output: GeneratorOutputType = None
        input_args = {}
        if self.mock_output:
            output = GeneratorOutput(data=self.mock_output_data)
        else:
            if self.teacher_mode and not isinstance(self, BackwardEngine):
                if not self._teacher:
                    if DEBUG_MODE:
                        print(
                            f"unwrapped_prompt_kwargs: {unwrapped_prompt_kwargs}, model_kwargs: {model_kwargs}"
                        )
                        print(f"names: {self.name}")
                    raise ValueError("Teacher generator is not set.")
                log.info(f"Using teacher: {self._teacher}")
                input_args = {
                    "prompt_kwargs": compose_model_kwargs(
                        self._teacher.prompt_kwargs, unwrapped_prompt_kwargs
                    ),
                    "model_kwargs": compose_model_kwargs(
                        self._teacher.model_kwargs, model_kwargs
                    ),
                }
                output = self._teacher.call(**input_args, id=id)
            else:
                input_args = {
                    "prompt_kwargs": compose_model_kwargs(
                        self.prompt_kwargs, unwrapped_prompt_kwargs
                    ),
                    "model_kwargs": compose_model_kwargs(
                        self.model_kwargs, model_kwargs
                    ),
                }
                # printc(f"input_args: {input_args}", color="red")

                output = self.call(**input_args, id=id)
                if not isinstance(output, GeneratorOutput):
                    raise ValueError(
                        f"Output should be of type GeneratorOutput, got {type(output)}"
                    )
        # 2. Generate a Parameter object from the output
        combined_prompt_kwargs = compose_model_kwargs(self.prompt_kwargs, prompt_kwargs)
        # if self.data_map_func is None:
        #     self.set_data_map_func()

        predecessors = [
            p for p in combined_prompt_kwargs.values() if isinstance(p, Parameter)
        ]

        log.debug(f"Predecessors: {predecessors} for generator {self.name}")

        def data_to_prompt_map_fn(data: Parameter) -> str:
            data: GeneratorOutput = data.data
            # if data.data is not None:
            #     return data.data
            if data.error is not None:
                return f"Response: {data.raw_response} parsed with error: {data.error}"
            return f" {data.raw_response}"

        # TODO: all parameter should just wrap the whole output.
        # this is for training.
        param_data = output
        response: Parameter = OutputParameter(
            data=param_data,
            name=self.name + "_output",
            role_desc=f"Output from (llm) {self.name}",
            param_type=ParameterType.GENERATOR_OUTPUT,
            data_id=id,
            full_response=output,  # the data structure
            data_in_prompt=data_to_prompt_map_fn,
        )
        response.set_predecessors(predecessors)
        response.trace_forward_pass(
            input_args=input_args, full_response=output, id=self.id, name=self.name
        )
        # setattr(response, "full_response", output)
        # *** special to the generator ***
        response.trace_api_kwargs(api_kwargs=self._trace_api_kwargs)
        # attach the demo to the demo parameter
        # if self.tracing:
        demo_param = self.find_demo_parameter(combined_prompt_kwargs)

        if demo_param:
            if id is None:
                raise ValueError(
                    "ID is required for tracing. Please pass it to your Geneartor call."
                )

            demo = self.create_demo_data_instance(
                prompt_kwargs,
                output,
                id=id,
            )
            demo_param.add_dataclass_to_trace(demo, is_teacher=self.teacher_mode)
        else:
            log.debug(
                "No demo parameter found in the prompt_kwargs. You can not trace the demo data."
            )

        # **** end of the special to the generator ****

        # if not self.backward_engine:
        #     # self.set_backward_engine()
        #     log.debug(f"Backward engine: {self.backward_engine}")

        # attach a funtion to compute gradient for predecessors

        printc(f"disable_backward_engine config: {self._disable_backward_engine}")

        response.set_grad_fn(
            BackwardContext(
                backward_fn=self.backward,
                backward_engine=self.backward_engine,
                response=response,
                prompt_kwargs=prompt_kwargs,
                template=self.template,
                prompt_str=self.get_prompt(**combined_prompt_kwargs),
                disable_backward_engine=self._disable_backward_engine,
                id=id,
            )
        )
        return response

    def backward(
        self,
        response: Parameter,  # the output of the forward pass
        prompt_kwargs: Dict,
        template: str,
        prompt_str: str,
        backward_engine: Optional["Generator"] = None,
        id: Optional[str] = None,  # the id of the input
        disable_backward_engine: bool = False,
    ) -> Parameter:

        log.info(f"Generator: Backward: {response.name}")

        backward_pass_setup = (
            backward_engine.backward_pass_setup if backward_engine else None
        )
        printc(
            f"backward pass setup: {backward_pass_setup}, name: {self.name}",
            color="red",
        )

        children_params = response.predecessors
        is_intermediate_node = True
        if response.get_gradient_and_context_text().strip() == "":
            log.info(f"Generator: Backward: No gradient found for {response}.")

        # backward score to the demo parameter
        for pred in children_params:
            # if pred.requires_opt:
            if response.score is not None:
                pred.set_score(response.score)
            log.debug(
                f"backpropagate the score {response.score} to {pred.name}, is_teacher: {self.teacher_mode}"
            )
            if pred.param_type == ParameterType.DEMOS:
                # Accumulate the score to the demo
                pred.add_score_to_trace(
                    trace_id=id, score=response.score, is_teacher=self.teacher_mode
                )
                log.debug(f"Pred: {pred.name}, traces: {pred._traces}")

        # 1.backward for text-gradients
        if backward_engine:
            log.debug(
                f"Generator: Backward engine is set for the generator. {backward_engine}"
            )
            # if response.backward_engine_disabled:
            #     for pred in children_params:
            #         pred.backward_engine_disabled = True
            #     return

            all_pred_at_once = backward_pass_setup.all_pred_at_once

            if not all_pred_at_once:
                for pred in children_params:
                    if not pred.requires_opt or pred.param_type == ParameterType.DEMOS:
                        log.debug(
                            f"EvalFnToTextLoss: Skipping {pred} as it does not require optimization."
                        )
                        continue

                    self._backward_through_one_predecessor(
                        pred=pred,
                        response=response,
                        prompt_kwargs=prompt_kwargs,
                        # template=template,
                        backward_engine=backward_engine,
                        prompt_str=prompt_str,
                        backward_pass_setup=backward_pass_setup,
                        is_intermediate_node=is_intermediate_node,
                        disable_backward_engine=disable_backward_engine,
                    )
            else:
                backward = False
                for pred in children_params:
                    if pred.requires_opt and pred.param_type in [
                        ParameterType.PROMPT,
                        ParameterType.GENERATOR_OUTPUT,
                        ParameterType.RETRIEVER_OUTPUT,
                        ParameterType.OUTPUT,
                    ]:
                        backward = True
                        break
                if backward:
                    # 2nd approach, backward all that need opt at once.
                    self._backward_through_all_predecessors(
                        children_params=children_params,
                        response=response,
                        prompt_kwargs=prompt_kwargs,
                        template=template,
                        backward_engine=backward_engine,
                        prompt_str=prompt_str,
                        backward_pass_setup=backward_pass_setup,
                        is_intermediate_node=is_intermediate_node,
                    )
        else:
            log.debug("Backward engine is not set for the generator. No text gradient.")

    @staticmethod
    def _backward_through_all_predecessors(
        children_params: List[Parameter],
        response: Parameter,
        prompt_kwargs: Dict[str, str],
        backward_engine: "BackwardEngine",
        backward_pass_setup: BackwardPassSetup,
        is_intermediate_node: bool = False,
    ):
        parser = JsonParser()
        # instruction and objective is the same for all the children
        instruction_str, objective_str = None, None

        # 1. Generate the conversation input and output
        input_prompt_kwargs = {
            k: v.get_prompt_data() if isinstance(v, Parameter) else v
            for k, v in prompt_kwargs.items()
        }

        print(f"gt: {response.get_gt()}")

        # TODO: pass all the parameters and even the templates
        conversation_prompt_kwargs = {
            "input_value": input_prompt_kwargs,
            "llm_output": response.get_prompt_data(),
        }

        conversation_str = Prompt(
            prompt_kwargs=conversation_prompt_kwargs,
            template=LLM_CONVERSATION_TEMPLATE,
        )()

        all_pred_info = Prompt(
            prompt_kwargs={"variables": [p.get_param_info() for p in children_params]},
            template=ALL_PRED_INFO,
        )()

        printc(f"all_pred_info: {all_pred_info}")

        conv_ins_template = None  # CONVERSATION_START_INSTRUCTION_BASE
        obj_ins_template = OBJECTIVE_INSTRUCTION_BASE
        if is_intermediate_node:  # TODO: this will always be true
            conv_ins_template = CONVERSATION_START_INSTRUCTION_CHAIN
            obj_ins_template = OBJECTIVE_INSTRUCTION_CHAIN
            response_gradient = response.get_gradients_str()
            # response_gradient = response.get_gradients_component_schema()
            # response_gradient = response.get_gradients_component_schema(
            #     skip_correct_sample=False
            # )
            if not response_gradient:
                raise ValueError(
                    f"Generator: No gradient found for {response}. Please check the response."
                )

        # replace variable and peers with all_pred_info

        instruction_str = Prompt(
            template=conv_ins_template,
            prompt_kwargs={
                "variable_and_peers_info": all_pred_info,
                "conversation_str": conversation_str,
            },
        )()
        objective_str = Prompt(
            template=obj_ins_template,
            prompt_kwargs={
                "response_desc": response.role_desc,
                "response_gradient": response_gradient,
                "instruction_to_backward_engine": response.instruction_to_backward_engine,
            },
        )()

        backward_engine_prompt_kwargs = {
            "conversation_sec": instruction_str,
            "objective_instruction_sec": objective_str,
            "output_format_str": OUTPUT_INSTRUCTION,
        }

        backward_engine_prompt_str = backward_engine.get_prompt(
            **backward_engine_prompt_kwargs
        )
        # print(f"Backward engine prompt: {backward_engine_prompt_str}")

        gradient_output: GeneratorOutput = None
        response_gradient_list = [""] * len(children_params)
        if (
            backward_pass_setup.compute_grad_for_errors_only
            and response.score is not None
            and float(response.score)
            > backward_pass_setup.threshold_score_to_compute_grad_for_errors
        ):
            manual_response_1 = f"Eval score: {response.score}. No noticeable error."
            response_gradient_list = [manual_response_1] * len(children_params)
            raw_response = str(response_gradient_list)
            gradient_output = GeneratorOutput(
                data=response_gradient_list, raw_response=raw_response
            )
        else:

            gradient_output: GeneratorOutput = backward_engine(
                prompt_kwargs=backward_engine_prompt_kwargs
            )
            if not isinstance(gradient_output, GeneratorOutput):
                raise ValueError(
                    f"Generator: Backward Engine should return a GeneratorOutput. Got {gradient_output} instead."
                )

            # parse the list of gradients

            try:
                response_gradient_list = parser.call(gradient_output.data)
            except Exception as e:
                log.error(f"Error parsing the response_gradient_list: {e}")
                failure_message = backward_engine.failure_message_to_optimizer(
                    gradient_output
                )
                if failure_message:
                    response_gradient_list = [failure_message] * len(children_params)
                printc(f"failure_message: {failure_message}", color="red")

        print(f"gradient list: {response_gradient_list}")

        # computes gradient for each prompt predecessor
        for i, pred in enumerate(children_params):
            if not pred.requires_opt or pred.param_type == ParameterType.DEMOS:
                log.debug(
                    f"Generator: Skipping {pred} as it does not require optimization."
                )
                continue

            gradient_data = (
                response_gradient_list[i]
                if response_gradient_list and len(response_gradient_list) > i
                else "Failed to get the gradient."
            )

            var_gradient = Gradient(
                data=gradient_data,
                data_id=response.data_id,
                score=response.score,  # add score to gradient
                from_response=response,
                to_pred=pred,
            )
            var_gradient.add_context(
                GradientContext(
                    input_output=conversation_str,
                    response_desc=response.role_desc,
                    variable_desc=pred.role_desc,  # the only difference for each pred
                )
            )
            var_gradient.add_prompt(backward_engine_prompt_str)
            pred.add_gradient(var_gradient)
            if response.score is not None:
                pred.set_score(response.score)

    @staticmethod
    def _backward_through_one_predecessor(
        pred: Parameter,
        response: Parameter,
        prompt_kwargs: Dict[str, str],
        backward_engine: "BackwardEngine",
        prompt_str: str,
        backward_pass_setup: BackwardPassSetup,
        is_intermediate_node: bool = False,
        disable_backward_engine: bool = False,
    ):
        """Creating gradient/textual feedback for prompt type parameters."""
        if not pred.requires_opt:
            if response.score is not None:
                pred.set_score(response.score)
            log.debug(
                f"Generator: Skipping {pred} as it does not require optimization."
            )
            return

        if pred.check_if_already_computed_gradient_respect_to(response.id):
            log.debug(
                f"Generator: Skipping {pred} as the gradient is already computed."
            )

            return

        if backward_engine is None:
            log.error(
                "EvalFnToTextLoss: backward_engine is required for text prompt optimization."
            )
            raise ValueError(
                "EvalFnToTextLoss: backward_engine is required for text prompt optimization."
            )

        instruction_str, objective_str = None, None

        # 1. Generate the conversation string
        input_prompt_kwargs = {
            k: v.get_prompt_data() if isinstance(v, Parameter) else v
            for k, v in prompt_kwargs.items()
        }

        conversation_prompt_kwargs = {
            "input_value": input_prompt_kwargs,
            "llm_output": response.get_prompt_data(),
            "gt": response.get_gt(),
        }

        conversation_str = Prompt(
            prompt_kwargs=conversation_prompt_kwargs,
            template=LLM_CONVERSATION_TEMPLATE,
        )()

        variable_dict = pred.get_param_info()

        peers = [p.get_param_info() for p in pred.peers]

        variable_and_peers_info = Prompt(
            prompt_kwargs={"variable": variable_dict, "peers": peers},
            template=VARIABLE_AND_PEERS_INFO,
        )()

        # generator is almost always intermediate node
        conv_ins_template = None  # CONVERSATION_START_INSTRUCTION_BASE
        obj_ins_template = OBJECTIVE_INSTRUCTION_BASE
        if is_intermediate_node:  # TODO: this will always be true
            conv_ins_template = CONVERSATION_START_INSTRUCTION_CHAIN
            obj_ins_template = OBJECTIVE_INSTRUCTION_CHAIN
            response_gradient = response.get_gradients_str()
            # response_gradient = response.get_gradients_component_schema()
            if not response_gradient:
                raise ValueError(
                    f"Generator: No gradient found for {response}. Please check the response. pred: {pred}"
                )
        predecessors = [
            pred.get_param_info()
            for pred in response.predecessors
            if pred not in pred.peers
        ]
        instruction_str = Prompt(
            template=conv_ins_template,
            prompt_kwargs={
                "variable_and_peers_info": variable_and_peers_info,
                "conversation_str": conversation_str,
                "predecessors": predecessors,
            },
        )()
        log.info(f"Conversation start instruction base str: {instruction_str}")
        objective_str = Prompt(
            template=obj_ins_template,
            prompt_kwargs={
                "response_desc": response.role_desc,
                "response_gradient": response_gradient,
                "instruction_to_backward_engine": pred.instruction_to_backward_engine,
            },
        )()

        backward_engine_prompt_kwargs = {
            "conversation_sec": instruction_str,
            "objective_instruction_sec": objective_str,
        }
        backward_engine_prompt_str = backward_engine.get_prompt(
            **backward_engine_prompt_kwargs
        )
        # print(f"Backward engine prompt: {backward_engine_prompt_str}")
        gradient_value = None
        if not disable_backward_engine:
            printc("doing backward engine")
            gradient_output: GeneratorOutput = None
            if (
                backward_pass_setup.compute_grad_for_errors_only
                and response.score is not None
                and float(response.score)
                > backward_pass_setup.threshold_score_to_compute_grad_for_errors
            ):
                log.debug(
                    f"EvalFnToTextLoss: Skipping {pred} as the score is high enough."
                )
                # TODO: plus score descriptions
                manual_response = f"Eval score: {response.score}. No noticeable error."
                gradient_output = GeneratorOutput(
                    data=manual_response, raw_response=manual_response
                )
            else:

                gradient_output: GeneratorOutput = backward_engine(
                    prompt_kwargs=backward_engine_prompt_kwargs
                )
                prompt_str = backward_engine.get_prompt(**backward_engine_prompt_kwargs)
                printc(f"Backward engine prompt: {prompt_str}")
                if not isinstance(gradient_output, GeneratorOutput):
                    raise ValueError(
                        f"Generator: Backward Engine should return a GeneratorOutput. Got {gradient_output} instead."
                    )
            printc(f"Backward engine gradient: {gradient_output}")

            # USE this to trace each node's input and output, all nodes can be visualized
            log.info(
                f"Generator Backward Engine Prompt: {backward_engine.get_prompt( **backward_engine_prompt_kwargs)}"
            )
            gradient_value = (
                gradient_output.data
                or backward_engine.failure_message_to_optimizer(gradient_output)
            )
        var_gradient = Gradient(
            data=gradient_value,
            data_id=response.data_id,
            score=response.score,  # add score to gradient
            from_response=response,
            to_pred=pred,
        )
        # Component-level input and output.
        var_gradient.add_context(
            GradientContext(
                input_output=conversation_str,
                response_desc=response.role_desc,
                variable_desc=pred.role_desc,  # parameter_desc
            )
        )
        var_gradient.add_prompt(backward_engine_prompt_str)
        pred.add_gradient(var_gradient)
        if response.score is not None:
            pred.set_score(response.score)

    def _run_callbacks(
        self,
        output: GeneratorOutput,
        input: Dict,
        prompt_kwargs: Dict,
        model_kwargs: Dict,
    ):
        self.trigger_callbacks(
            "on_complete",
            output=output,
            input=input,
            prompt_kwargs=prompt_kwargs,
            model_kwargs=model_kwargs,
        )
        if output.error:
            self.trigger_callbacks(
                "on_failure",
                output=output,
                input=input,
                prompt_kwargs=prompt_kwargs,
                model_kwargs=model_kwargs,
            )
        else:
            self.trigger_callbacks(
                "on_success",
                output=output,
                input=input,
                prompt_kwargs=prompt_kwargs,
                model_kwargs=model_kwargs,
            )

    def call(
        self,
        prompt_kwargs: Optional[Dict] = {},  # supports both str and parameter value
        model_kwargs: Optional[Dict] = {},
        use_cache: Optional[bool] = None,
        id: Optional[str] = None,
    ) -> GeneratorOutputType:
        r"""
        Call the model_client by formatting prompt from the prompt_kwargs,
        and passing the combined model_kwargs to the model client.
        """
        if self.mock_output:
            return GeneratorOutput(data=self.mock_output_data, id=id)

        log.debug(f"prompt_kwargs: {prompt_kwargs}")
        log.debug(f"model_kwargs: {model_kwargs}")

        api_kwargs = self._pre_call(prompt_kwargs, model_kwargs)

        log.debug(f"api_kwargs: {api_kwargs}")
        output: GeneratorOutputType = None
        # call the model client

        completion = None
        use_cache = use_cache if use_cache is not None else self._use_cache
        try:
            completion = self._model_client_call(
                api_kwargs=api_kwargs, use_cache=use_cache
            )
        except Exception as e:
            log.error(f"Error calling the model: {e}")
            output = GeneratorOutput(error=str(e), id=id)
        # process the completion
        if completion is not None:
            try:
                output = self._post_call(completion)
            except Exception as e:
                log.error(f"Error processing the output: {e}")
                output = GeneratorOutput(
                    raw_response=str(completion), error=str(e), id=id
                )

        # User only need to use one of them, no need to use them all.
        output.id = id
        self._run_callbacks(
            output,
            input=api_kwargs,
            prompt_kwargs=prompt_kwargs,
            model_kwargs=model_kwargs,
        )

        log.info(f"output: {output}")
        self._trace_api_kwargs = api_kwargs  # tracing
        return output

    # TODO: training is not supported in async call yet
    async def acall(
        self,
        prompt_kwargs: Optional[Dict] = {},
        model_kwargs: Optional[Dict] = {},
        use_cache: Optional[bool] = None,
        id: Optional[str] = None,
    ) -> GeneratorOutputType:
        r"""Async call the model with the input and model_kwargs.

        :warning::
            Training is not supported in async call yet.
        """
        log.info(f"prompt_kwargs: {prompt_kwargs}")
        log.info(f"model_kwargs: {model_kwargs}")

        api_kwargs = self._pre_call(prompt_kwargs, model_kwargs)
        output: GeneratorOutputType = None
        # call the model client
        completion = None

        try:
            completion = await self.model_client.acall(
                api_kwargs=api_kwargs, model_type=self.model_type
            )
        except Exception as e:
            log.error(f"Error calling the model: {e}")
            output = GeneratorOutput(error=str(e))

        if completion:
            try:
                output = self._post_call(completion)
            except Exception as e:
                log.error(f"Error processing the output: {e}")
                output = GeneratorOutput(raw_response=str(completion), error=str(e))

        log.info(f"output: {output}")
        self._run_callbacks(
            output,
            input=api_kwargs,
            prompt_kwargs=prompt_kwargs,
            model_kwargs=model_kwargs,
        )
        self._trace_api_kwargs = api_kwargs  # tracing
        return output

    def __call__(self, *args, **kwargs) -> Union[GeneratorOutputType, Any]:
        if self.training:
            log.debug("Training mode")
            return self.forward(*args, **kwargs)
        else:
            log.debug("Inference mode")
            return self.call(*args, **kwargs)

    def _extra_repr(self) -> str:
        # Create the string for model_kwargs
        s = f"model_kwargs={self.model_kwargs}, "

        # Create the string for trainable prompt_kwargs
        prompt_kwargs_repr = [
            k
            for k, v in self.prompt_kwargs.items()
            if isinstance(v, Parameter) and v.requires_opt
        ]

        s += f"trainable_prompt_kwargs={prompt_kwargs_repr}"
        s += f", prompt={self.prompt}"
        return s

    def to_dict(self) -> Dict[str, Any]:
        r"""Convert the generator to a dictionary."""
        # TODO: exclude default functions
        return super().to_dict()

    @staticmethod
    def failure_message_to_backward_engine(
        gradient_response: GeneratorOutput,
    ) -> Optional[str]:
        response_value = None
        if gradient_response.error or not gradient_response.data:
            response_value = f"Error: {gradient_response.error}, Raw response: {gradient_response.raw_response}"
        return response_value


class BackwardEngine(Generator):  # it is a generator with defaule template

    __doc__ = """A Generator with a default template for the backward pass in auto-differentiation.

    As a component, the forward pass is simply the same as the call method.
    So it will always return GeneratorOutputType instead of Parameter.

    If you want to customize the template, you can create your own backward engine.
    Yet, we will forever keep the training mode to False for the backward engine.
    This is achieved by making forward the same as call.
    """

    def __init__(self, **kwargs):
        if kwargs is None:
            kwargs = {}
        kwargs["template"] = FEEDBACK_ENGINE_TEMPLATE

        super().__init__(**kwargs)
        self.name = "BackwardEngine"
        self.teacher_mode = False

    def call(self, **kwargs) -> GeneratorOutputType:
        r"""Catch the rate limit error and raise it."""
        output = super().call(**kwargs)
        if output and output.error is not None and "429" in output.error:
            raise ValueError(f"Error in the backward engine: {output.error}")
        return output

    def forward(self, **kwargs):
        r"""Forward pass for the backward engine."""
        return self.call(**kwargs)

    @staticmethod
    def failure_message_to_optimizer(
        gradient_response: GeneratorOutput,
    ) -> Optional[str]:
        gradient_value_data = None
        if gradient_response.error or not gradient_response.data:
            gradient_value_data = f"The backward engine failed to compute the gradient. Raw response: {gradient_response.raw_response}, Error: {gradient_response.error}"

        return gradient_value_data


def create_teacher_generator(
    student: Generator,
    model_client: ModelClient,
    model_kwargs: Dict[str, Any],
    template: Optional[str] = None,
) -> Generator:
    r"""Create a teacher generator from the student generator.

    Note:
        Teacher generator will have no parameters.
        If you want to keep it to be the same as the student, just create one each time your student has been updated.
        Or else, task.parameters will list teacher parameters.

    Args:
        student (Generator): The student generator.
        model_client (ModelClient): The model client to use for the teacher generator.
        model_kwargs (Dict[str, Any]): The model kwargs to pass to the model client.
        name (str, optional): The name of the teacher generator. Defaults to "teacher".

    Returns:
        Generator: The teacher generator.
    """
    kwargs = student._kwargs.copy()
    kwargs["model_client"] = model_client
    kwargs["model_kwargs"] = model_kwargs
    if template:
        kwargs["template"] = template
    kwargs["name"] = f"{student.name}_teacher"

    prompt_kwargs_str: Dict[str, str] = {}
    for key, p in kwargs["prompt_kwargs"].items():
        if isinstance(p, Parameter):
            prompt_kwargs_str[key] = str(p.data)
        else:
            prompt_kwargs_str[key] = p
    kwargs["prompt_kwargs"] = prompt_kwargs_str
    teacher = Generator(
        **kwargs,
    )
    return teacher


if __name__ == "__main__":
    # test the generator with backward engine
    # TODO: move this to external local tests before packaging
    from adalflow.components.model_client import (
        GroqAPIClient,
        OpenAIClient,
        GoogleGenAIClient,
        AnthropicAPIClient,
    )
    from adalflow.utils import setup_env
    from adalflow.core.model_client import ModelClient

    setup_env()
    # log = get_logger(level="DEBUG")
    llama3_model = {
        "model_client": GroqAPIClient(),
        "model_kwargs": {
            "model": "llama-3.1-8b-instant",
        },
    }
    gpt_3_model = {
        "model_client": OpenAIClient(),
        "model_kwargs": {
            "model": "gpt-3.5-turbo",
        },
    }
    gemini_model = {
        "model_client": GoogleGenAIClient(),
        "model_kwargs": {
            "model": "gemini-1.0-pro",
        },
    }
    claude_model = {
        "model_client": AnthropicAPIClient(),
        "model_kwargs": {
            "model": "claude-3-opus-20240229",
            "max_tokens": 100,
        },
    }
    from adalflow.tracing.generator_call_logger import GeneratorCallLogger
    from functools import partial

    # setup the logger
    call_logger = GeneratorCallLogger(save_dir="traces")

    def on_complete(output, input, prompt_kwargs, model_kwargs, logger_call: Callable):
        logger_call(
            output=output,
            input=input,
            prompt_kwargs=prompt_kwargs,
            model_kwargs=model_kwargs,
        )

    for model in [llama3_model, gpt_3_model, gemini_model, claude_model]:
        generator = Generator(**model)

        teacher = create_teacher_generator(generator, **claude_model)

        call_logger.register_generator("generator", "generator_call")
        # setup the callback
        logger_call = partial(call_logger.log_call, name="generator")
        generator.register_callback(
            "on_complete", partial(on_complete, logger_call=logger_call)
        )

        output = generator(
            prompt_kwargs={
                "input_str": "Hello, world!",
            }
        )
        break

    # test the backward engine
    # TODO: test ollama and transformer client to update the change



from typing import List, Tuple
import numpy as np


def confidence_interval(
    judgements: List[float], confidence: float = 0.95
) -> Tuple[float, float]:
    """
    Calculate the confidence interval for a list of binary judgments.

    Args:
        judgements (List[float]): List of binary judgments (1/0).
        confidence (float): Confidence level (default 0.95).

    Returns:
        tuple: Lower and upper bounds of the confidence interval.
    """
    if not isinstance(judgements, list):
        raise ValueError("judgements must be a list")

    if not all(isinstance(j, (int, float)) for j in judgements):
        raise ValueError("judgements must contain only integers or floats")

    if not isinstance(confidence, (int, float)):
        raise ValueError("confidence must be a number")

    if not 0 < confidence < 1:
        raise ValueError("confidence must be between 0 and 1")

    # Step 1: Calculate the mean
    mean_score = np.mean(judgements)

    # Step 2: Calculate the standard error (SE)
    standard_error = np.std(judgements, ddof=1) / np.sqrt(len(judgements))

    # Step 3: Use the Z-critical value for the confidence interval based on confidence level
    z_critical = np.percentile(
        np.random.normal(0, 1, 1000000), 100 * (1 - (1 - confidence) / 2)
    )

    # Step 4: Calculate the margin of error (MoE)
    margin_of_error = z_critical * standard_error

    # Step 5: Calculate the confidence interval
    # Confidence interval (clipped to [0, 1])
    lower_bound = max(0, mean_score - margin_of_error)
    upper_bound = min(1, mean_score + margin_of_error)

    return (lower_bound, upper_bound)


def longest_common_substring(s1: str, s2: str) -> str:
    """
    Find the longest common substring between two strings.
    """
    # Create a matrix to store lengths of longest common suffixes of substrings
    # Initialize all values to 0
    m, n = len(s1), len(s2)
    lcs_matrix = [[0] * (n + 1) for _ in range(m + 1)]

    # Variable to store the length of the longest common substring
    longest_length = 0
    # Variable to store the ending index of the longest common substring in s1
    ending_index_s1 = 0

    # Build the matrix in a bottom-up manner
    for i in range(1, m + 1):
        for j in range(1, n + 1):
            if s1[i - 1] == s2[j - 1]:
                lcs_matrix[i][j] = lcs_matrix[i - 1][j - 1] + 1
                if lcs_matrix[i][j] > longest_length:
                    longest_length = lcs_matrix[i][j]
                    ending_index_s1 = i - 1
            else:
                lcs_matrix[i][j] = 0

    # The longest common substring
    longest_common_substring = s1[
        ending_index_s1 - longest_length + 1 : ending_index_s1 + 1
    ]

    return longest_common_substring


if __name__ == "__main__":
    # Example binary judgments (True/False as 1/0)
    judgements = [1, 1, 0, 1, 0, 1, 1]  # Convert to 1/0
    score_range = confidence_interval(judgements, confidence=0.96)
    print(score_range)

    # Example longest common substring
    s1 = "abcdfghijk"
    s2 = "abedfghxyz"
    cs = longest_common_substring(s1, s2)
    print(cs)
